try:
    from dotenv import load_dotenv
    load_dotenv()
except ImportError:
    pass

import json
import os
from typing import Optional

import requests
from validator_collection import validators

from highcharts_core import errors, constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core.options import HighchartsOptions
from highcharts_core.options.data import Data


class ExportServer(HighchartsMeta):
    """Class that provides methods for interacting with the Highcharts
    `Export Server <https://github.com/highcharts/node-export-server>`_.

    .. note::

      By default, the :class:`ExportServer` class operates using the Highcharts-provided
      export server. If you wish to use your own (or a custom) export server, you can
      configure the class using either the :meth:`url <ExportServer.url>`,
      :meth:`port <ExportServer.port>`, and
      :meth:`path <ExportServer.path>` properties explicitly or by setting
      the ``HIGHCHARTS_EXPORT_SERVER_DOMAIN`, ``HIGHCHARTS_EXPORT_SERVER_PORT``, or
      ``HIGHCHARTS_EXPORT_SERVER_PATH`` environment variables.

    """

    def __init__(self, **kwargs):
        self._url = None
        self._port = None
        self._path = None
        self._options = None
        self._format_ = None
        self._scale = None
        self._width = None
        self._callback = None
        self._constructor = None
        self._use_base64 = None
        self._no_download = None
        self._async_rendering = None
        self._global_options = None
        self._data_options = None
        self._custom_code = None

        self.protocol = kwargs.get('protocol',
                                   os.getenv('HIGHCHARTS_EXPORT_SERVER_PROTOCOL',
                                             'https'))
        self.domain = kwargs.get('domain', os.getenv('HIGHCHARTS_EXPORT_SERVER_DOMAIN',
                                                     'export.highcharts.com'))
        self.port = kwargs.get('port', os.getenv('HIGHCHARTS_EXPORT_SERVER_PORT',
                                                 ''))
        self.path = kwargs.get('path', os.getenv('HIGHCHARTS_EXPORT_SERVER_PATH',
                                                 ''))
        self.options = kwargs.get('options', None)
        self.format_ = kwargs.get('format_', kwargs.get('type', 'png'))
        self.scale = kwargs.get('scale', 1)
        self.width = kwargs.get('width', None)
        self.callback = kwargs.get('callback', None)
        self.constructor = kwargs.get('constructor', 'Chart')
        self.use_base64 = kwargs.get('use_base64', False)
        self.no_download = kwargs.get('no_download', False)
        self.async_rendering = kwargs.get('async_rendering', False)
        self.global_options = kwargs.get('global_options', None)
        self.data_options = kwargs.get('data_options', None)
        self.custom_code = kwargs.get('custom_code', None)

        super().__init__(**kwargs)

    @property
    def protocol(self) -> Optional[str]:
        """The protocol over which the Highcharts for Python library should communicate
        with the :term:`Export Server`. Accepts either ``'https'`` or ``'http'``. Defaults
        to the ``HIGHCHARTS_EXPORT_SERVER_PROTOCOL`` environment variable if present,
        otherwise falls back to default of ``'https'``.

        .. tip::

          This property is set automatically by the ``HIGHCHARTS_EXPORT_SERVER_PROTOCOL``
          environment variable, if present.

        .. warning::

          If set to :obj:`None <python:None>`, will fall back to the
          ``HIGHCHARTS_EXPORT_SERVER_PROTOCOL`` value if available, and the Highsoft-
          provided server (``'export.highcharts.com'``) if not.

        :rtype: :class:`str <python:str>`
        """
        return self._protocol

    @protocol.setter
    def protocol(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            value = os.getenv('HIGHCHARTS_EXPORT_SERVER_PROTOCOL', 'https')

        value = value.lower()
        if value not in ['https', 'http']:
            raise errors.HighchartsUnsupportedProtocolError(f'protocol expects either '
                                                            f'"https" or "http". '
                                                            f'Received: "{value}"')

        self._protocol = value
        self._url = None

    @property
    def domain(self) -> Optional[str]:
        """The domain where the :term:`Export Server` can be found. Defaults to the
        Highsoft-provided Export Server at ``'export.highcharts.com'``, unless over-ridden
        by the ``HIGHCHARTS_EXPORT_SERVER_DOMAIN`` environment variable.

        .. tip::

          This property is set automatically by the ``HIGHCHARTS_EXPORT_SERVER_DOMAIN``
          environment variable, if present.

        .. warning::

          If set to :obj:`None <python:None>`, will fall back to the
          ``HIGHCHARTS_EXPORT_SERVER_DOMAIN`` value if available, and the Highsoft-
          provided server (``'export.highcharts.com'``) if not.

        :rtype: :class:`str <pythoon:str>`
        """
        return self._domain

    @domain.setter
    def domain(self, value):
        value = validators.domain(value, allow_empty = True)
        if not value:
            value = os.getenv('HIGHCHARTS_EXPORT_SERVER_DOMAIN',
                              'export.highcharts.com')
        self._domain = value
        self._url = None

    @property
    def port(self) -> Optional[int]:
        """The port on which the :term:`Export Server` can be found. Defaults to
        :obj:`None <python:None>` (for the Highsoft-provided export server), unless
        over-ridden by the ``HIGHCHARTS_EXPORT_SERVER_PORT`` environment variable.

        .. tip::

          This property is set automatically by the ``HIGHCHARTS_EXPORT_SERVER_PORT``
          environment variable, if present.

        .. warning::

          If set to :obj:`None <python:None>`, will fall back to the
          ``HIGHCHARTS_EXPORT_SERVER_PORT`` value if available. If unavailable, will
          revert to :obj:`None <python:None>`.

        :rtype: :class:`str <pythoon:str>`
        """
        return self._port

    @port.setter
    def port(self, value):
        if value or value == 0:
            value = validators.integer(value,
                                       allow_empty = True,
                                       minimum = 0,
                                       maximum = 65536)
        else:
            value = os.getenv('HIGHCHARTS_EXPORT_SERVER_PORT', None)

        self._port = value
        self._url = None

    @property
    def path(self) -> Optional[str]:
        """The path (at the :meth:`ExportServer.url`) where the :term:`Export Server` can
        be reached. Defaults to :obj:`None <python:None>` (for the Highsoft-provided
        export server), unless over-ridden by the ``HIGHCHARTS_EXPORT_SERVER_PATH``
        environment variable.

        .. tip::

          This property is set automatically by the ``HIGHCHARTS_EXPORT_SERVER_PATH``
          environment variable, if present.

        .. warning::

          If set to :obj:`None <python:None>`, will fall back to the
          ``HIGHCHARTS_EXPORT_SERVER_PATH`` value if available. If unavailable, will
          revert to :obj:`None <python:None>`.

        :rtype: :class:`str <pythoon:str>`
        """
        return self._path

    @path.setter
    def path(self, value):
        value = validators.path(value, allow_empty = True)
        if value is None:
            value = os.getenv('HIGHCHARTS_EXPORT_SERVER_PATH', None)

        self._path = value
        self._url = None

    @property
    def url(self) -> Optional[str]:
        """The fully-formed URL for the :term:`Export Server`, consisting of a
        :meth:`protocol <ExportServer.protocol>`, a :meth:`domain <ExportServer.domain>`,
        and optional :meth:`port <ExportServer.port>` and
        :meth:`path <ExportServer.path>`.

        .. note::

          If explicitly set, will override the values in related properties:

            * :meth:`protocol <ExportServer.protocol>`,
            * :meth:`domain <ExportServer.domain>`,
            * :meth:`port <ExportServer.port>`, and
            * :meth:`path <ExportServer.path>`

        :rtype: :class:`str <python:str>`
        """
        if self._url:
            return self._url
        else:
            return_value = f'{self.protocol}://{self.domain}'
            if self.port is not None:
                return_value += f':{self.port}/'
            if self.path is not None:
                return_value += self.path

            return return_value

    @url.setter
    def url(self, value):
        value = validators.url(value, allow_empty = True)
        if not value:
            self.protocol = None
            self.domain = None
            self.port = None
            self.path = None
        else:
            original_value = value
            self.protocol = value[:value.index(':')]

            protocol = self.protocol + '://'
            value = value.replace(protocol, '')

            no_port = False
            try:
                end_of_domain = value.index(':')
                self.domain = value[:end_of_domain]
            except ValueError:
                no_port = True
                try:
                    end_of_domain = value.index('/')
                    self.domain = value[:end_of_domain]
                except ValueError:
                    self.domain = value

            domain = self.domain + '/'
            if domain in value:
                value = value.replace(domain, '')
            elif self.domain in value:
                value = value.replace(self.domain, '')

            if value and no_port:
                if value.startswith('/'):
                    self.path = value[1:]
                else:
                    self.path = value
            else:
                if value.startswith(':'):
                    start_of_port = 1
                else:
                    start_of_port = 0
                try:
                    end_of_port = value.index('/')
                except ValueError:
                    end_of_port = None

                if end_of_port:
                    self.port = value[start_of_port:end_of_port]
                else:
                    self.port = value[start_of_port:]

                port = f':{self.port}'
                value = value.replace(port, '')
                if value.startswith('/'):
                    self.path = value[1:]
                elif value:
                    self.path = value
                else:
                    self.path = None

            self._url = original_value

    @property
    def options(self) -> Optional[HighchartsOptions]:
        """The :class:`HighchartsOptions` which should be applied to render the exported
        chart. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`HighchartsOptions` or :obj:`None <pythoN:None>`
        """
        return self._options

    @options.setter
    @class_sensitive(HighchartsOptions)
    def options(self, value):
        self._options = value

    @property
    def format_(self) -> Optional[str]:
        """The format in which the exported chart should be returned. Defaults to
        ``'png'``.

        Accepts:

          * ``'png'``
          * ``'jpeg'``
          * ``'pdf'``
          * ``'svg'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format_

    @format_.setter
    def format_(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._format_ = None
        else:
            value = value.lower()
            if value not in ['png', 'jpeg', 'pdf', 'svg']:
                raise errors.HighchartsUnsupportedExportTypeError(
                    f'format_ expects either '
                    f'"png", "jpeg", "pdf", or '
                    f'"svg". Received: {value}'
                )
            self._format_ = value

    @property
    def scale(self) -> Optional[int | float]:
        """The scale factor by which the exported chart image should be scaled. Defaults
        to ``1``.

        .. tip::

          Use this setting to improve resolution when exporting PNG or JPEG images. For
          example, setting ``.scale = 2`` on a chart whose width is 600px will produce
          an image with a width of 1200px.

        .. warning::

          If :meth:`width <ExportServer.width>` is explicitly set, this setting will be
          overridden.

        :rtype: numeric
        """
        return self._scale

    @scale.setter
    def scale(self, value):
        value = validators.numeric(value,
                                   allow_empty = True,
                                   minimum = 0)
        if not value:
            value = 1

        self._scale = value

    @property
    def width(self) -> Optional[int | float]:
        """The width that the exported chart should have. Defaults to
        :obj:`None <python:None>`.

        .. warning::

          If explicitly set, this setting will override
          :meth:`scale <ExportServer.scale>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        value = validators.numeric(value,
                                   allow_empty = True,
                                   minimum = 0)
        if not value:
            value = None

        self._width = value

    @property
    def callback(self) -> Optional[CallbackFunction]:
        """A JavaScript function to execute in the (JavaScript) Highcharts constructor.

        .. note::

          This setting is equivalent to providing the :meth:`Chart.callback` setting.

        :rtype: :class:`CallbackFunction` or :obj:`None <pythoN:None>`
        """
        return self._callback

    @callback.setter
    @class_sensitive(CallbackFunction)
    def callback(self, value):
        self._callback = value

    @property
    def constructor(self) -> Optional[str]:
        """The (JavaScript) constructor to use when generating the exported chart.
        Defaults to :obj:`None <python:None>`.

        Accepts:

          * ``'Chart'``
          * ``'Stock'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._constructor

    @constructor.setter
    def constructor(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._constructor = None
        else:
            if value not in ['Chart', 'Stock']:
                raise errors.HighchartsUnsupportedConstructorError(f'constructor expects '
                                                                   f'either "Chart" or '
                                                                   f'"Stock", but '
                                                                   f'received: "{value}"')

            self._constructor = value

    @property
    def use_base64(self) -> bool:
        """If ``True``, returns the exported chart in base64 encoding. If ``False``,
        returns the exported chart in binary. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>`
        """
        return self._use_base64

    @use_base64.setter
    def use_base64(self, value):
        self._use_base64 = bool(value)

    @property
    def no_download(self) -> bool:
        """If ``True``, will not send attachment headers in the HTTP response when
        exporting a chart. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>`
        """
        return self._no_download

    @no_download.setter
    def no_download(self, value):
        self._no_download = bool(value)

    @property
    def async_rendering(self) -> bool:
        """If ``True``, will delay the (server-side) rendering of the exported chart
        until all scripts, functions, and event handlers provided have been executed
        and the (JavaScript) method ``highexp.done()`` is called. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>`
        """
        return self._async_rendering

    @async_rendering.setter
    def async_rendering(self, value):
        self._async_rendering = bool(value)

    @property
    def global_options(self) -> Optional[HighchartsOptions]:
        """The global options which will be passed to the (JavaScript)
        ``Highcharts.setOptions()`` method, and which will be applied to the exported
        chart. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`HighchartsOptions`
        """
        return self._global_options

    @global_options.setter
    @class_sensitive(HighchartsOptions)
    def global_options(self, value):
        self._global_options = value

    @property
    def data_options(self) -> Optional[Data]:
        """Configuration of data options to add data to the chart from sources like CSV.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`Data` or :obj:`None <python:None>`
        """
        return self._data_options

    @data_options.setter
    @class_sensitive(Data)
    def data_options(self, value):
        self._data_options = value

    @property
    def custom_code(self) -> Optional[CallbackFunction]:
        """When :meth:`data_options <ExportServer.data_options>` is not
        :obj:`None <python:None>`, this (JavaScript) callback function is executed after
        the data options are applied. The only argument it receives is the complete
        set of :class:`HighchartsOptions` (as a JS literal object), which will be passed
        to the Highcharts constructor on return. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._custom_code

    @custom_code.setter
    @class_sensitive(CallbackFunction)
    def custom_code(self, value):
        self._custom_code = value

    @classmethod
    def is_export_supported(cls, options) -> bool:
        """Evaluates whether the Highcharts Export Server supports exporting the series types in ``options``.
        
        :rtype: :class:`bool <python:bool>`
        """
        if not isinstance(options, HighchartsOptions):
            return False

        if not options.series:
            return True

        series_types = [x.type for x in options.series]
        for item in series_types:
            if item in constants.EXPORT_SERVER_UNSUPPORTED_SERIES_TYPES:
                return False
            
        return True

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        url = as_dict.get('url', None)
        protocol = None
        domain = None
        port = None
        path = None
        if not url:
            protocol = as_dict.get('protocol', None)
            domain = as_dict.get('domain', None)
            port = as_dict.get('port', None)
            path = as_dict.get('path', None)

        kwargs = {
            'options': as_dict.get('options', None),
            'format_': as_dict.get('type', as_dict.get('format_', 'png')),
            'scale': as_dict.get('scale', 1),
            'width': as_dict.get('width', None),
            'callback': as_dict.get('callback', None),
            'constructor': as_dict.get('constructor',
                                       None) or as_dict.get('constr', None),
            'use_base64': as_dict.get('use_base64', None) or as_dict.get('b64', False),
            'no_download': as_dict.get('noDownload',
                                       None) or as_dict.get('no_download', None),
            'async_rendering': as_dict.get('asyncRendering',
                                           False) or as_dict.get('async_rendering',
                                                                 False),
            'global_options': as_dict.get('global_options',
                                          None) or as_dict.get('globalOptions',
                                                               None),
            'data_options': as_dict.get('data_options',
                                        None) or as_dict.get('dataOptions', None),
            'custom_code': as_dict.get('custom_code',
                                       None) or as_dict.get('customCode', None)
        }
        if url:
            kwargs['url'] = url
        if protocol:
            kwargs['protocol'] = protocol
        if domain:
            kwargs['domain'] = domain
        if port:
            kwargs['port'] = port
        if path:
            kwargs['path'] = path

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'url': self.url,
            'options': self.options,
            'type': self.format_,
            'scale': self.scale,
            'width': self.width,
            'callback': self.callback,
            'constr': self.constructor,
            'b64': self.use_base64,
            'noDownload': self.no_download,
            'asyncRendering': self.async_rendering,
            'globalOptions': self.global_options,
            'dataOptions': self.data_options,
            'customCode': self.custom_code
        }

        return untrimmed

    def request_chart(self,
                      filename = None,
                      auth_user = None,
                      auth_password = None,
                      timeout = 3,
                      **kwargs):
        """Execute a request against the export server based on the configuration in the
        instance.

        :param filename: The name of the file where the exported chart should (optionally)
          be persisted. Defaults to :obj:`None <python:None>`.
        :type filename: Path-like or :obj:`None <python:None>`

        :param auth_user: The username to use to authenticate against the
          Export Server, using :term:`basic authentication`. Defaults to
          :obj:`None <python:None>`.
        :type auth_user: :class:`str <python:str>` or :obj:`None <python:None>`

        :param auth_password: The password to use to authenticate against the Export
          Server (using :term:`basic authentication`). Defaults to
          :obj:`None <python:None>`.
        :type auth_password: :class:`str <python:str>` or :obj:`None <python:None>`

        :param timeout: The number of seconds to wait before issuing a timeout error.
          The timeout check is passed if bytes have been received on the socket in less
          than the ``timeout`` value. Defaults to ``3``.
        :type timeout: numeric or :obj:`None <python:None>`

        .. note::

          All other keyword arguments are as per the :class:`ExportServer` constructor
          :meth:`ExportServer.__init__() <highcharts_core.headless_export.ExportServer.__init__>`

        :returns: The exported chart image, either as a :class:`bytes <python:bytes>`
          binary object or as a base-64 encoded string (depending on the
          :meth:`use_base64 <ExportServer.use_base64>` property).
        :rtype: :class:`bytes <python:bytes>` or :class:`str <python:str>`
        """
        self.options = kwargs.get('options', self.options)
        self.format_ = kwargs.get('format_', kwargs.get('type', self.format_))
        self.scale = kwargs.get('scale', self.scale)
        self.width = kwargs.get('width', self.width)
        self.callback = kwargs.get('callback', self.callback)
        self.constructor = kwargs.get('constructor', self.constructor)
        self.use_base64 = kwargs.get('use_base64', self.use_base64)
        self.no_download = kwargs.get('no_download', self.no_download)
        self.async_rendering = kwargs.get('async_rendering', self.async_rendering)
        self.global_options = kwargs.get('global_options', self.global_options)
        self.data_options = kwargs.get('data_options', self.data_options)
        self.custom_code = kwargs.get('custom_code', self.custom_code)

        missing_details = []
        if not self.options:
            missing_details.append('options')
        if not self.format_:
            missing_details.append('format_')
        if not self.constructor:
            missing_details.append('constructor')
        if not self.url:
            missing_details.append('url')

        if missing_details:
            raise errors.HighchartsMissingExportSettingsError(
                f'Unable to export a chart.'
                f'ExportServer was missing '
                f' following settings: '
                f'{missing_details}'
            )

        basic_auth = None
        if auth_user and auth_password:
            basic_auth = requests.HTTPBasicAuth(auth_user, auth_password)

        payload = {
            'infile': 'HIGHCHARTS FOR PYTHON: REPLACE WITH OPTIONS',
            'type': self.format_,
            'scale': self.scale,
            'constr': self.constructor,
            'b64': self.use_base64,
            'noDownload': self.no_download,
            'asyncRendering': self.async_rendering
        }
        if self.width:
            payload['width'] = self.width
        if self.callback:
            payload['callback'] = 'HIGHCHARTS FOR PYTHON: REPLACE WITH CALLBACK'
        if self.global_options:
            payload['globalOptions'] = 'HIGHCHARTS FOR PYTHON: REPLACE WITH GLOBAL'
        if self.data_options:
            payload['dataOptions'] = 'HIGHCHARTS FOR PYTHON: REPLACE WITH DATA'
        if self.custom_code:
            payload['customCode'] = 'HIGHCHARTS FOR PYTHON: REPLACE WITH CUSTOM'

        as_json = json.dumps(payload)
        
        if not self.is_export_supported(self.options):
            raise errors.HighchartsUnsupportedExportError('The Highcharts Export Server currently only supports '
                                                          'exports from Highcharts (Javascript) v.10. You are '
                                                          'using a series type introduced in v.11. Sorry, but '
                                                          'that functionality is still forthcoming.')
        
        options_as_json = self.options.to_json()
        if isinstance(options_as_json, bytes):
            options_as_str = str(options_as_json, encoding = 'utf-8')
        else:
            options_as_str = options_as_json

        as_json = as_json.replace('"HIGHCHARTS FOR PYTHON: REPLACE WITH OPTIONS"',
                                  options_as_str)
        if self.callback:
            callback_as_json = self.callback.to_json()
            if isinstance(callback_as_json, bytes):
                callback_as_str = str(callback_as_json, encoding = 'utf-8')
            else:
                callback_as_str = callback_as_json
            as_json = as_json.replace('"HIGHCHARTS FOR PYTHON: REPLACE WITH CALLBACK"',
                                      callback_as_str)
        if self.global_options:
            global_as_json = self.global_options.to_json()
            if isinstance(global_as_json, bytes):
                global_as_str = str(global_as_json, encoding = 'utf-8')
            else:
                global_as_str = global_as_json
            as_json = as_json.replace('"HIGHCHARTS FOR PYTHON: REPLACE WITH GLOBAL"',
                                      global_as_str)
        if self.data_options:
            data_as_json = self.data_options.to_json()
            if isinstance(data_as_json, bytes):
                data_as_str = str(data_as_json, encoding = 'utf-8')
            else:
                data_as_str = data_as_json
            as_json = as_json.replace('"HIGHCHARTS FOR PYTHON: REPLACE WITH DATA"',
                                      data_as_str)
        if self.custom_code:
            code_as_json = self.custom_code.to_json()
            if isinstance(code_as_json, bytes):
                code_as_str = str(code_as_json, encoding = 'utf-8')
            else:
                code_as_str = code_as_json
            as_json = as_json.replace('"HIGHCHARTS FOR PYTHON: REPLACE WITH CUSTOM"',
                                      code_as_str)

        result = requests.post(self.url,
                               data = as_json,
                               headers = { 'Content-Type': 'application/json' },
                               auth = basic_auth,
                               timeout = timeout)

        result.raise_for_status()

        if filename and self.format_ != 'svg':
            with open(filename, 'wb') as file_:
                file_.write(result.content)
        elif filename and self.format_ == 'svg':
            content = str(result.content, encoding = 'utf-8')
            with open(filename, 'wt') as file_:
                file_.write(content)

        return result.content

    @classmethod
    def get_chart(cls,
                  filename = None,
                  auth_user = None,
                  auth_password = None,
                  timeout = 3,
                  **kwargs):
        """Produce an exported chart image.

        :param filename: The name of the file where the exported chart should (optionally)
          be persisted. Defaults to :obj:`None <python:None>`.
        :type filename: Path-like or :obj:`None <python:None>`

        :param auth_user: The username to use to authenticate against the
          Export Server, using :term:`basic authentication`. Defaults to
          :obj:`None <python:None>`.
        :type auth_user: :class:`str <python:str>` or :obj:`None <python:None>`

        :param auth_password: The password to use to authenticate against the Export
          Server (using :term:`basic authentication`). Defaults to
          :obj:`None <python:None>`.
        :type auth_password: :class:`str <python:str>` or :obj:`None <python:None>`

        :param timeout: The number of seconds to wait before issuing a timeout error.
          The timeout check is passed if bytes have been received on the socket in less
          than the ``timeout`` value. Defaults to ``3``.
        :type timeout: numeric or :obj:`None <python:None>`

        .. note::

          All other keyword arguments are as per the :class:`ExportServer` constructor
          :meth:`ExportServer.__init__() <highcharts_core.headless_export.ExportServer.__init__>`

        :returns: The exported chart image, either as a :class:`bytes <python:bytes>`
          binary object or as a base-64 encoded string (depending on the ``use_base64``
          keyword argument).
        :rtype: :class:`bytes <python:bytes>` or :class:`str <python:str>`
        """
        instance = cls(**kwargs)

        exported_chart = instance.request_chart(filename = filename,
                                                auth_user = auth_user,
                                                auth_password = auth_password,
                                                timeout = timeout)

        return exported_chart



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class Loading(HighchartsMeta):
    """The loading options control the appearance of the loading screen that covers
    the plot area on chart operations.

    This screen only appears after an explicit call to ``chart.showLoading()`` in the
    browser. It is a utility for developers to communicate to the end user that
    something is going on, for example while retrieving new data via an XHR
    connection.

    .. hint::

      The "Loading..." text itself is **not** part of this configuration object, but
      is instead part of the :meth:`.language <Options.language>` configuration.

    """

    def __init__(self, **kwargs):
        self._hide_duration = None
        self._label_style = None
        self._show_duration = None
        self._style = None

        self.hide_duration = kwargs.get('hide_duration', None)
        self.label_style = kwargs.get('label_style', None)
        self.show_duration = kwargs.get('show_duration', None)
        self.style = kwargs.get('style', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'loading'

    @property
    def hide_duration(self) -> Optional[int]:
        """The duration in milliseconds of the fade out effect. Defaults to
        ``100``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._hide_duration

    @hide_duration.setter
    def hide_duration(self, value):
        self._hide_duration = validators.integer(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def label_style(self) -> Optional[str]:
        """CSS styles applied to the loading label's ``<span>``. Defaults to
        ``'{"fontWeight": "bold", "position": "relative", "top": "45%"}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._label_style

    @label_style.setter
    def label_style(self, value):
        self._label_style = validators.string(value, allow_empty = True)

    @property
    def show_duration(self) -> Optional[int]:
        """The duration in milliseconds of the fade in effect. Defaults to
        ``100``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._show_duration

    @show_duration.setter
    def show_duration(self, value):
        self._show_duration = validators.integer(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def style(self) -> Optional[str]:
        """CSS styles for the loading screen that covers the plot area. Defaults to
        ``'{"position": "absolute", "backgroundColor": "#ffffff", "opacity": 0.5, "textAlign": "center"}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'hide_duration': as_dict.get('hideDuration', None),
            'label_style': as_dict.get('labelStyle', None),
            'show_duration': as_dict.get('showDuration', None),
            'style': as_dict.get('style', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'hideDuration': self.hide_duration,
            'labelStyle': self.label_style,
            'showDuration': self.show_duration,
            'style': self.style
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.position import Position
from highcharts_core.utility_classes.events import PointEvents


class ConnectorOptions(HighchartsMeta):
    """Options for the connector in the Series on point feature."""

    def __init__(self, **kwargs):
        self._dashstyle = None
        self._stroke = None
        self._width = None

        self.dashstyle = kwargs.get('dashstyle', None)
        self.stroke = kwargs.get('stroke', None)
        self.width = kwargs.get('width', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.series.onPoint.connectorOptions'

    @property
    def dashstyle(self) -> Optional[str]:
        """Name of the dash style to use for the connector.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dashstyle

    @dashstyle.setter
    def dashstyle(self, value):
        if not value:
            self._dashstyle = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'dashstyle expects a recognized value'
                                                  f', but received: {value}')
            self._dashstyle = value

    @property
    def stroke(self) -> Optional[str]:
        """The color of the collector line.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stroke

    @stroke.setter
    def stroke(self, value):
        self._stroke = validators.string(value, allow_empty = True)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """Pixel width of the connector line. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'dashstyle': as_dict.get('dashstyle', None) or as_dict.get('dashStyle', None),
            'stroke': as_dict.get('stroke', None),
            'width': as_dict.get('width', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'dashstyle': self.dashstyle,
            'stroke': self.stroke,
            'width': self.width
        }

        return untrimmed


class OnPointOptions(HighchartsMeta):
    """Options for the Series on point feature, which is currently only supported by
    ``pie`` and ``sunburst`` chargs."""

    def __init__(self, **kwargs):
        self._connector_options = None
        self._id = None
        self._position = None

        self.connector_options = kwargs.get('connector_options', None)
        self.id = kwargs.get('id', None)
        self.position = kwargs.get('position', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.series.onPoint'

    @property
    def connector_options(self) -> Optional[ConnectorOptions]:
        """Options for the connector in the Series on point feature.

        :rtype: :class:`ConnectorOptions` or :obj:`None <python:None>`
        """
        return self._connector_options

    @connector_options.setter
    @class_sensitive(ConnectorOptions)
    def connector_options(self, value):
        self._connector_options = value

    @property
    def id(self) -> Optional[str]:
        """The ``id`` of the point that we connect the series to.

        .. note::

          Only points with a given ``plot_x`` and ``plot_y`` values and map points are
          valid.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = validators.string(value, allow_empty = True)

    @property
    def position(self) -> Optional[Position]:
        """Options allowing to set a position and an offset of the series in the Series
        on point feature.

        :rtype: :class:`Position` or :obj:`None <python:None>`
        """
        return self._position

    @position.setter
    @class_sensitive(Position)
    def position(self, value):
        self._position = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'connector_options': as_dict.get('connectorOptions', None),
            'id': as_dict.get('id', None),
            'position': as_dict.get('position', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'connectorOptions': self.connector_options,
            'id': self.id,
            'position': self.position
        }

        return untrimmed


class Point(HighchartsMeta):
    """Properties for each single point."""

    def __init__(self, **kwargs):
        self._events = None

        self.events = kwargs.get('events', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.series.point'

    @property
    def events(self) -> Optional[PointEvents]:
        """Events for each single point.

        :rtype: :class:`PointEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(PointEvents)
    def events(self, value):
        self._events = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'events': as_dict.get('events', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'events': self.events
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta


class AnnotationAnimation(HighchartsMeta):
    """Configuraiton settings that apply to the animation of an annotation."""

    def __init__(self, **kwargs):
        self._defer = None

        self.defer = kwargs.get('defer', None)

    @property
    def defer(self) -> Optional[int | float | Decimal]:
        """The animation delay time in milliseconds. Defaults to
        :obj:`None <python:None>`.

        Set to ``0`` to render the annotation immediately.

        Set to :obj:`None <python:None>` to inherit the defer time from
        :meth:`Series.animation.defer`.

        :returns: The animation delay time in milliseconds.
        :rtype: numeric or :obj:`None <python:None>`

        :raises ValueError: if set to a negative number
        """
        return self._defer

    @defer.setter
    def defer(self, value):
        self._defer = validators.numeric(value,
                                         allow_empty = True,
                                         minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'defer': as_dict.get('defer', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'defer': self.defer
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.series.treegraph import TreegraphSeries as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'data': [
          {
            'colorValue': 2,
            'dataLabels': {
              'align': 'center',
              'allowOverlap': True,
              'animation': {
                  'defer': 5
              },
              'backgroundColor': {
                  'linearGradient': {
                      'x1': 0.123,
                      'x2': 0.234,
                      'y1': 0.345,
                      'y2': 0.456
                  },
                  'stops': [
                      [0.12, '#999'],
                      [0.34, '#fff']
                  ]
              },
              'borderColor': '#999999',
              'borderRadius': 24,
              'borderWidth': 1,
              'className': 'some-class-name',
              'color': '#000000',
              'crop': True,
              'defer': False,
              'enabled': True,
              'filter': {
                  'operator': '>=',
                  'property': 'some_property',
                  'value': 123
              },
              'format': 'some format',
              'formatter': """function() { return true; }""",
              'inside': True,
              'nullFormat': 'some format',
              'nullFormatter': """function() { return true; }""",
              'overflow': 'none',
              'padding': 12,
              'position': 'center',
              'rotation': 0,
              'shadow': False,
              'shape': 'rect',
              'style': 'style goes here',
              'useHTML': False,
              'verticalAlign': 'top',
              'x': 10,
              'y': 20,
              'z': 0
            },
            'dragDrop': {
                'draggableX': True,
                'draggableY': True,
                'dragHandle': {
                    'className': 'draghandle-classname-goes-here',
                    'color': '#ccc',
                    'cursor': 'alias',
                    'lineColor': '#ddd',
                    'lineWidth': 2,
                    'pathFormatter': """function() { return true; }""",
                    'zIndex': 10
                },
                'dragMaxX': 3456,
                'dragMaxY': 6532,
                'dragMinX': 123,
                'dragMinY': 321,
                'dragPrecisionX': 5,
                'dragPrecisionY': 5,
                'dragSensitivity': 2,
                'groupBy': 'some-property-name',
                'guideBox': {
                    'default': {
                        'className': 'some-classname-goes-here',
                        'color': '#999',
                        'cursor': 'pointer',
                        'lineColor': '#ccc',
                        'lineWidth': 2,
                        'zIndex': 100
                    }
                },
                'liveRedraw': True
            },
            'drilldown': 'some-id-goes-here',
            'parent': 'some-id-goes-here',
          }
      ]
    }, None),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('series/treegraph/01.js', False, None),
    ('series/treegraph/02.js', False, None),

    ('series/treegraph/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('series/treegraph/error-02.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('series/treegraph/01.js', True, None),
    ('series/treegraph/02.js', True, None),

    ('series/treegraph/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('series/treegraph/error-02.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.plot_options.arcdiagram import ArcDiagramOptions
from highcharts_core.options.plot_options.area import AreaOptions
from highcharts_core.options.plot_options.area import AreaRangeOptions
from highcharts_core.options.plot_options.area import AreaSplineOptions
from highcharts_core.options.plot_options.area import AreaSplineRangeOptions
from highcharts_core.options.plot_options.bar import BarOptions
from highcharts_core.options.plot_options.bellcurve import BellCurveOptions
from highcharts_core.options.plot_options.boxplot import BoxPlotOptions
from highcharts_core.options.plot_options.bubble import BubbleOptions
from highcharts_core.options.plot_options.bullet import BulletOptions
from highcharts_core.options.plot_options.bar import ColumnOptions
from highcharts_core.options.plot_options.bar import ColumnPyramidOptions
from highcharts_core.options.plot_options.bar import ColumnRangeOptions
from highcharts_core.options.plot_options.bar import CylinderOptions
from highcharts_core.options.plot_options.dependencywheel import DependencyWheelOptions
from highcharts_core.options.plot_options.dumbbell import DumbbellOptions
from highcharts_core.options.plot_options.boxplot import ErrorBarOptions
from highcharts_core.options.plot_options.funnel import FunnelOptions
from highcharts_core.options.plot_options.funnel import Funnel3DOptions
from highcharts_core.options.plot_options.gauge import GaugeOptions
from highcharts_core.options.plot_options.heatmap import HeatmapOptions
from highcharts_core.options.plot_options.histogram import HistogramOptions
from highcharts_core.options.plot_options.item import ItemOptions
from highcharts_core.options.plot_options.area import LineOptions
from highcharts_core.options.plot_options.dumbbell import LollipopOptions
from highcharts_core.options.plot_options.networkgraph import NetworkGraphOptions
from highcharts_core.options.plot_options.organization import OrganizationOptions
from highcharts_core.options.plot_options.packedbubble import PackedBubbleOptions
from highcharts_core.options.plot_options.pareto import ParetoOptions
from highcharts_core.options.plot_options.pictorial import PictorialOptions
from highcharts_core.options.plot_options.pie import PieOptions
from highcharts_core.options.plot_options.polygon import PolygonOptions
from highcharts_core.options.plot_options.pyramid import PyramidOptions
from highcharts_core.options.plot_options.pyramid import Pyramid3DOptions
from highcharts_core.options.plot_options.sankey import SankeyOptions
from highcharts_core.options.plot_options.scatter import ScatterOptions
from highcharts_core.options.plot_options.scatter import Scatter3DOptions
from highcharts_core.options.plot_options.series import SeriesOptions
from highcharts_core.options.plot_options.gauge import SolidGaugeOptions
from highcharts_core.options.plot_options.spline import SplineOptions
from highcharts_core.options.plot_options.area import StreamGraphOptions
from highcharts_core.options.plot_options.sunburst import SunburstOptions
from highcharts_core.options.plot_options.heatmap import TilemapOptions
from highcharts_core.options.plot_options.timeline import TimelineOptions
from highcharts_core.options.plot_options.treegraph import TreegraphOptions
from highcharts_core.options.plot_options.treemap import TreemapOptions
from highcharts_core.options.plot_options.pie import VariablePieOptions
from highcharts_core.options.plot_options.bar import VariwideOptions
from highcharts_core.options.plot_options.vector import VectorOptions
from highcharts_core.options.plot_options.venn import VennOptions
from highcharts_core.options.plot_options.bar import WaterfallOptions
from highcharts_core.options.plot_options.bar import WindBarbOptions
from highcharts_core.options.plot_options.wordcloud import WordcloudOptions
from highcharts_core.options.plot_options.bar import XRangeOptions


class PlotOptions(HighchartsMeta):
    """A wrapper object for configurations applied to each series type.

    The config objects for each series can also be overridden for each series item as
    given in the series array.

    Configuration options for the series are given in three levels:

      * Options for all series in a chart are given in the
        :meth:`series <PlotOptions.series>` property.
      * Options for all series of a specific type are given in the corresponding
        property for that type, for example
        :meth:`plot_options.line <PlotOptions.line>`.
      * Finally, options for one single series are given in the
        :meth:`Options.series <Options.series>` array.

    """

    def __init__(self, **kwargs):
        self._arcdiagram = None
        self._area = None
        self._arearange = None
        self._areaspline = None
        self._areasplinerange = None
        self._bar = None
        self._bellcurve = None
        self._boxplot = None
        self._bubble = None
        self._bullet = None
        self._column = None
        self._columnpyramid = None
        self._columnrange = None
        self._cylinder = None
        self._dependencywheel = None
        self._dumbbell = None
        self._errorbar = None
        self._funnel = None
        self._funnel_3d = None
        self._gauge = None
        self._heatmap = None
        self._histogram = None
        self._item = None
        self._line = None
        self._lollipop = None
        self._networkgraph = None
        self._organization = None
        self._packedbubble = None
        self._pareto = None
        self._pictorial = None
        self._pie = None
        self._polygon = None
        self._pyramid = None
        self._pyramid_3d = None
        self._sankey = None
        self._scatter = None
        self._scatter_3d = None
        self._series = None
        self._solidgauge = None
        self._spline = None
        self._streamgraph = None
        self._sunburst = None
        self._tilemap = None
        self._timeline = None
        self._treegraph = None
        self._treemap = None
        self._variablepie = None
        self._variwide = None
        self._vector = None
        self._venn = None
        self._waterfall = None
        self._windbarb = None
        self._wordcloud = None
        self._xrange = None

        for attribute in dir(self):
            if attribute.startswith('_') and not attribute.startswith('__'):
                non_private_name = attribute[1:]
                setattr(self, non_private_name, kwargs.get(non_private_name, None))

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions'

    @property
    def arcdiagram(self) -> Optional[ArcDiagramOptions]:
        """Arc diagram series is a chart drawing style in which the vertices of the chart
        are positioned along a line on the Euclidean plane and the edges are drawn as a
        semicircle in one of the two half-planes delimited by the line, or as smooth
        curves formed by sequences of semicircles.

        .. figure:: ../../../_static/arcdiagram-example.png
          :alt: Arc Diagram Example Chart
          :align: center

        :rtype: :class:`ArcDiagramOptions` or :obj:`None <python:None>`
        """
        return self._arcdiagram

    @arcdiagram.setter
    @class_sensitive(ArcDiagramOptions)
    def arcdiagram(self, value):
        self._arcdiagram = value

    @property
    def area(self) -> Optional[AreaOptions]:
        """General options to apply to all Area series types.

        .. figure:: ../../../_static/area-example.png
          :alt: Area Example Chart
          :align: center

        :rtype: :class:`AreaOptions` or :obj:`None <python:None>`
        """
        return self._area

    @area.setter
    @class_sensitive(AreaOptions)
    def area(self, value):
        self._area = value

    @property
    def arearange(self) -> Optional[AreaRangeOptions]:
        """General options to apply to all AreaRange series types. The area range series
        is a carteseian series with higher and lower values for each point along an X
        axis, where the area between the values is shaded.

        .. figure:: ../../../_static/arearange-example.png
          :alt: AreaRange Example Chart
          :align: center

        :rtype: :class:`AreaRangeOptions` or :obj:`None <python:None>`
        """
        return self._arearange

    @arearange.setter
    @class_sensitive(AreaRangeOptions)
    def arearange(self, value):
        self._arearange = value

    @property
    def areaspline(self) -> Optional[AreaSplineOptions]:
        """General options to apply to all AreaSpline series types. The area spline series
        is an area series where the graph between the points is smoothed into a spline.

        .. figure:: ../../../_static/areaspline-example.png
          :alt: AreaSpline Example Chart
          :align: center

        :rtype: :class:`AreaSplineOptions` or :obj:`None <python:None>`
        """
        return self._areaspline

    @areaspline.setter
    @class_sensitive(AreaSplineOptions)
    def areaspline(self, value):
        self._areaspline = value

    @property
    def areasplinerange(self) -> Optional[AreaSplineRangeOptions]:
        """General options to apply to all AreaSplineRange series types. The area spline
        range series is a carteseian series type with higher and lower Y values along an X
        axis. The area inside the range is colored, and the graph outlining the area is a
        smoothed spline.

        :rtype: :class:`AreaSplineRangeOptions` or :obj:`None <python:None>`
        """
        return self._areasplinerange

    @areasplinerange.setter
    @class_sensitive(AreaSplineRangeOptions)
    def areasplinerange(self, value):
        self._areasplinerange = value

    @property
    def bar(self) -> Optional[BarOptions]:
        """General options to apply to all Bar series types. A bar series is a special
        type of column series where the columns are horizontal.

        .. figure:: ../../../_static/bar-example.png
          :alt: Bar Example Chart
          :align: center

        :rtype: :class:`BarOptions` or :obj:`None <python:None>`
        """
        return self._bar

    @bar.setter
    @class_sensitive(BarOptions)
    def bar(self, value):
        self._bar = value

    @property
    def bellcurve(self) -> Optional[BellCurveOptions]:
        """General options to apply to all Bell Curve series types.

        A bell curve is an areaspline series which represents the probability density
        function of the normal distribution. It calculates mean and standard deviation of
        the base series data and plots the curve according to the calculated parameters.

        .. figure:: ../../../_static/bellcurve-example.png
          :alt: Bell Curve Example Chart
          :align: center

        :rtype: :class:`BellCurveOptions` or :obj:`None <python:None>`
        """
        return self._bellcurve

    @bellcurve.setter
    @class_sensitive(BellCurveOptions)
    def bellcurve(self, value):
        self._bellcurve = value

    @property
    def boxplot(self) -> Optional[BoxPlotOptions]:
        """General options to apply to all Box Plot series types.

        A box plot is a convenient way of depicting groups of data through their
        five-number summaries:

          * the smallest observation (sample minimum),
          * lower quartile (Q1),
          * median (Q2),
          * upper quartile (Q3), and
          * largest observation (sample maximum).

        .. figure:: ../../../_static/boxplot-example.png
          :alt: Box Plot Example Chart
          :align: center

        :rtype: :class:`BoxPlotOptions` or :obj:`None <python:None>`
        """
        return self._boxplot

    @boxplot.setter
    @class_sensitive(BoxPlotOptions)
    def boxplot(self, value):
        self._boxplot = value

    @property
    def bubble(self) -> Optional[BubbleOptions]:
        """General options to apply to all Bubble series types.

        A bubble series is a three dimensional series type where each point renders an X,
        Y, and Z value. Each points is drawn as a bubble where the position along the X
        and Y axes mark the X and Y values, and the size of the bubble relates to the Z
        value.

        .. figure:: ../../../_static/bubble-example.png
          :alt: Bubble Example Chart
          :align: center

        :rtype: :class:`BubbleOptions` or :obj:`None <python:None>`
        """
        return self._bubble

    @bubble.setter
    @class_sensitive(BubbleOptions)
    def bubble(self, value):
        self._bubble = value

    @property
    def bullet(self) -> Optional[BulletOptions]:
        """General options to apply to all Bullet series types.

        A bullet graph is a variation of a bar graph. The bullet graph features a single
        measure, compares it to a target, and displays it in the context of qualitative
        ranges of performance that could be set using :meth:`YAxis.plot_bands`.

        .. figure:: ../../../_static/bullet-example.png
          :alt: Bullet Example Chart
          :align: center

        :rtype: :class:`BulletOptions` or :obj:`None <python:None>`
        """
        return self._bullet

    @bullet.setter
    @class_sensitive(BulletOptions)
    def bullet(self, value):
        self._bullet = value

    @property
    def column(self) -> Optional[ColumnOptions]:
        """General options to apply to all Column series types.

        Column series display one column per value along an X axis.

        .. figure:: ../../../_static/column-example.png
          :alt: Column Example Chart
          :align: center

        :rtype: :class:`ColumnOptions` or :obj:`None <python:None>`
        """
        return self._column

    @column.setter
    @class_sensitive(ColumnOptions)
    def column(self, value):
        self._column = value

    @property
    def columnpyramid(self) -> Optional[ColumnPyramidOptions]:
        """General options to apply to all Column Pyramid series types.

        Column Pyramid series display one pyramid per value along an X axis.

        .. hint::

          To display horizontal pyramids, set :meth:`Chart.inverted` to ``True``.

        .. tabs::

          .. tab:: Standard

            .. figure:: ../../../_static/columnpyramid-example.png
              :alt: ColumnPyramid Example Chart
              :align: center

          .. tab:: Stacked

            .. figure:: ../../../_static/columnpyramid-example-stacked.png
              :alt: Stacked Column Pyramid Example Chart
              :align: center

          .. tab:: Stacked + Inverted

            .. figure:: ../../../_static/columnpyramid-example-stacked-horizontal.png
              :alt: Stacked and Inverted Column Pyramid Example Chart
              :align: center

        :rtype: :class:`ColumnPyramidOptions` or :obj:`None <python:None>`
        """
        return self._columnpyramid

    @columnpyramid.setter
    @class_sensitive(ColumnPyramidOptions)
    def columnpyramid(self, value):
        self._columnpyramid = value

    @property
    def columnrange(self) -> Optional[ColumnRangeOptions]:
        """General options to apply to all Column Range series types.

        The column range is a cartesian series type with higher and lower Y values along
        an X axis.

        .. hint::

          To display horizontal bars, set :meth:`Chart.inverted` to ``True``.

        .. tabs::

          .. tab:: Standard

            .. figure:: ../../../_static/columnrange-example.png
              :alt: ColumnRange Example Chart
              :align: center

          .. tab:: Horizontal

            .. figure:: ../../../_static/columnrange-example-horizontal.png
              :alt: Inverted Column Range Example Chart
              :align: center

        :rtype: :class:`ColumnRangeOptions` or :obj:`None <python:None>`
        """
        return self._columnrange

    @columnrange.setter
    @class_sensitive(ColumnRangeOptions)
    def columnrange(self, value):
        self._columnrange = value

    @property
    def cylinder(self) -> Optional[CylinderOptions]:
        """General options to apply to all Cylinder series types.

        A cylinder graph is a variation of a 3d column graph. The cylinder graph features
        cylindrical points.

        .. figure:: ../../../_static/cylinder-example.png
          :alt: Cylinder Example Chart
          :align: center

        :rtype: :class:`CylinderOptions` or :obj:`None <python:None>`
        """
        return self._cylinder

    @cylinder.setter
    @class_sensitive(CylinderOptions)
    def cylinder(self, value):
        self._cylinder = value

    @property
    def dependencywheel(self) -> Optional[DependencyWheelOptions]:
        """General options to apply to all Dependency Wheel series types.

        A dependency wheel chart is a type of flow diagram, where all nodes are laid out
        in a circle, and the flow between the are drawn as link bands.

        .. figure:: ../../../_static/dependencywheel-example.png
          :alt: Dependency Wheel Example Chart
          :align: center

        :rtype: :class:`DependencyWheelOptions` or :obj:`None <python:None>`
        """
        return self._dependencywheel

    @dependencywheel.setter
    @class_sensitive(DependencyWheelOptions)
    def dependencywheel(self, value):
        self._dependencywheel = value

    @property
    def dumbbell(self) -> Optional[DumbbellOptions]:
        """General options to apply to all Dumbbell series types.

        The dumbbell series is a cartesian series with higher and lower values for each
        point along an X axis, connected with a line between the values.

        .. figure:: ../../../_static/dumbbell-example.png
          :alt: Dumbbell Example Chart
          :align: center

        .. warning::

          Requires ``highcharts-more.js`` and ``modules/dumbbell.js`` be loaded
          client-side.

        :rtype: :class:`DumbbellOptions` or :obj:`None <python:None>`
        """
        return self._dumbbell

    @dumbbell.setter
    @class_sensitive(DumbbellOptions)
    def dumbbell(self, value):
        self._dumbbell = value

    @property
    def errorbar(self) -> Optional[ErrorBarOptions]:
        """General options to apply to all Error Bar series types.

        Error bars are a graphical representation of the variability of data and are used
        on graphs to indicate the error, or uncertainty in a reported measurement.

        .. figure:: ../../../_static/errorbar-example.png
          :alt: Error Bar Example Chart
          :align: center

        :rtype: :class:`ErrorBarOptions` or :obj:`None <python:None>`
        """
        return self._errorbar

    @errorbar.setter
    @class_sensitive(ErrorBarOptions)
    def errorbar(self, value):
        self._errorbar = value

    @property
    def funnel(self) -> Optional[FunnelOptions]:
        """General options to apply to all Funnel series types.

        Funnel charts are a type of chart often used to visualize stages in a sales
        project, where the top are the initial stages with the most clients.

        .. warning::

          Funnel charts require that the ``modules/funnel.js`` file is loaded client-side.

        .. figure:: ../../../_static/funnel-example.png
          :alt: Funnel Example Chart
          :align: center

        :rtype: :class:`FunnelOptions` or :obj:`None <python:None>`
        """
        return self._funnel

    @funnel.setter
    @class_sensitive(FunnelOptions)
    def funnel(self, value):
        self._funnel = value

    @property
    def funnel_3d(self) -> Optional[Funnel3DOptions]:
        """General options to apply to all Funnel 3D series types.

        A Funnel 3D chart is a three-dimensional version of funnel series type. Funnel
        charts are a type of chart often used to visualize stages in a sales project,
        where the top are the initial stages with the most clients.

        .. warning::

          Funnel 3D charts require that the following files are all loaded client-side:

            * ``highcharts-3d.js``,
            * ``cylinder.js`` and
            * ``funnel3d.js``

        .. figure:: ../../../_static/funnel_3d-example.png
          :alt: Funnel 3D Example Chart
          :align: center

        :rtype: :class:`FunnelOptions` or :obj:`None <python:None>`
        """
        return self._funnel_3d

    @funnel_3d.setter
    @class_sensitive(Funnel3DOptions)
    def funnel_3d(self, value):
        self._funnel_3d = value

    @property
    def gauge(self) -> Optional[GaugeOptions]:
        """General options to apply to all Gauge series types.

        Gauges are circular plots displaying one or more values with a dial pointing to
        values along the perimeter.

        .. figure:: ../../../_static/gauge-example.png
          :alt: Gauge Example Chart
          :align: center

        :rtype: :class:`GaugeOptions` or :obj:`None <python:None>`
        """
        return self._gauge

    @gauge.setter
    @class_sensitive(GaugeOptions)
    def gauge(self, value):
        self._gauge = value

    @property
    def heatmap(self) -> Optional[HeatmapOptions]:
        """General options to apply to all Heatmap series types.

        A heatmap is a graphical representation of data where the individual values
        contained in a matrix are represented as colors.

        .. warning::

          Heatmaps require that ``modules/heatmap`` is loaded client-side.

        .. figure:: ../../../_static/heatmap-example.png
          :alt: Heatmap Example Chart
          :align: center

        :rtype: :class:`HeatmapOptions` or :obj:`None <python:None>`
        """
        return self._heatmap

    @heatmap.setter
    @class_sensitive(HeatmapOptions)
    def heatmap(self, value):
        self._heatmap = value

    @property
    def histogram(self) -> Optional[HistogramOptions]:
        """General options to apply to all Histogram series types.

        A histogram is a column series which represents the distribution of the data set
        in the base series. Histogram splits data into bins and shows their frequencies.

        .. figure:: ../../../_static/histogram-example.png
          :alt: Histogram Example Chart
          :align: center

        :rtype: :class:`HistogramOptions` or :obj:`None <python:None>`
        """
        return self._histogram

    @histogram.setter
    @class_sensitive(HistogramOptions)
    def histogram(self, value):
        self._histogram = value

    @property
    def item(self) -> Optional[ItemOptions]:
        """General options to apply to all Item series types.

        An item chart is an infographic chart where a number of items are laid out in
        either a rectangular or circular pattern. It can be used to visualize counts
        within a group, or for the circular pattern, typically a parliament.

        The circular layout has much in common with a pie chart. Many of the item series
        options, like ``center``, ``size`` and data label positioning, are inherited from
        the :meth:`PlotOptions.pie` series and don't apply for rectangular layouts.

        .. tabs::

          .. tab:: Circular Item Chart

            .. figure:: ../../../_static/item-example-circular.png
              :alt: Circular Item Example Chart
              :align: center

          .. tab:: Rectangular Item Chart

            .. figure:: ../../../_static/item-example-rectangular.png
              :alt: Rectangular Item Example Chart
              :align: center

          .. tab:: Item Chart with Symbols

            .. figure:: ../../../_static/item-example-symbols.png
              :alt: Item Example Chart with Symbols
              :align: center

        :rtype: :class:`ItemOptions` or :obj:`None <python:None>`
        """
        return self._item

    @item.setter
    @class_sensitive(ItemOptions)
    def item(self, value):
        self._item = value

    @property
    def line(self) -> Optional[LineOptions]:
        """General options to apply to all Line series types.

        A line series displays information as a series of data points connected by
        straight line segments.

        .. figure:: ../../../_static/line-example.png
          :alt: Line Example Chart
          :align: center

        :rtype: :class:`LineOptions` or :obj:`None <python:None>`
        """
        return self._line

    @line.setter
    @class_sensitive(LineOptions)
    def line(self, value):
        self._line = value

    @property
    def lollipop(self) -> Optional[LollipopOptions]:
        """General options to apply to all Lollipop series types.

        The lollipop series is a carteseian series with a line anchored from the x axis
        and a dot at the end to mark the value.

        .. warning::

          Requires ``highcharts-more.js``, ``modules/dumbbell.js``, and
          ``modules/lollipop.js`` to be loaded client-side.

        .. figure:: ../../../_static/lollipop-example.png
          :alt: Lollipop Example Chart
          :align: center

        :rtype: :class:`LollipopOptions` or :obj:`None <python:None>`
        """
        return self._lollipop

    @lollipop.setter
    @class_sensitive(LollipopOptions)
    def lollipop(self, value):
        self._lollipop = value

    @property
    def networkgraph(self) -> Optional[NetworkGraphOptions]:
        """General options to apply to all Network Graph series types.

        A network graph is a type of relationship chart, where connnections (links)
        attract nodes (points) and other nodes repulse each other.

        .. figure:: ../../../_static/networkgraph-example.png
          :alt: NetworkGraph Example Chart
          :align: center

        :rtype: :class:`NetworkGraphOptions` or :obj:`None <python:None>`
        """
        return self._networkgraph

    @networkgraph.setter
    @class_sensitive(NetworkGraphOptions)
    def networkgraph(self, value):
        self._networkgraph = value

    @property
    def organization(self) -> Optional[OrganizationOptions]:
        """General options to apply to all Organization series types.

        An organization chart is a diagram that shows the structure of an organization and
        the relationships and relative ranks of its parts and positions.

        .. tabs::

          .. tab:: Standard Organization Chart

            .. figure:: ../../../_static/organization-example.png
              :alt: Organization Example Chart
              :align: center

          .. tab:: Horizontal Organization Chart

            .. figure:: ../../../_static/organization-example-horizontal.png
              :alt: Horizontal Organization Example Chart
              :align: center

        :rtype: :class:`OrganizationOptions` or :obj:`None <python:None>`
        """
        return self._organization

    @organization.setter
    @class_sensitive(OrganizationOptions)
    def organization(self, value):
        self._organization = value

    @property
    def packedbubble(self) -> Optional[PackedBubbleOptions]:
        """General options to apply to all Packed Bubble series types.

        A packed bubble series is a two dimensional series type, where each point renders
        a value in X, Y position. Each point is drawn as a bubble where the bubbles don't
        overlap with each other and the radius of the bubble relates to the value.

        .. tabs::

          .. tab:: Standard Packed Bubble

            .. figure:: ../../../_static/packedbubble-example.png
              :alt: Split Packed Bubble Example Chart
              :align: center

          .. tab:: Split Packed Bubble

            .. figure:: ../../../_static/packedbubble-example-split.png
              :alt: Split Packed Bubble Example Chart
              :align: center

        :rtype: :class:`PackedBubbleOptions` or :obj:`None <python:None>`
        """
        return self._packedbubble

    @packedbubble.setter
    @class_sensitive(PackedBubbleOptions)
    def packedbubble(self, value):
        self._packedbubble = value

    @property
    def pareto(self) -> Optional[ParetoOptions]:
        """General options to apply to all Pareto series types.

        A pareto diagram is a type of chart that contains both bars and a line graph,
        where individual values are represented in descending order by bars, and the
        cumulative total is represented by the line.

        .. figure:: ../../../_static/pareto-example.png
          :alt: Pareto Example Chart
          :align: center

        :rtype: :class:`ParetoOptions` or :obj:`None <python:None>`
        """
        return self._pareto

    @pareto.setter
    @class_sensitive(ParetoOptions)
    def pareto(self, value):
        self._pareto = value

    @property
    def pictorial(self) -> Optional[PictorialOptions]:
        """General options to apply to all Pictorial series types.

        A pictorial series uses vector images to represent the data, with the data's shape
        determined by the ``path`` parameter.

        .. figure:: ../../../_static/pictorial-example.png
          :alt: Pictorial Example Chart
          :align: center


        :rtype: :class:`PictorialOptions <highcharts_core.options.plot_options.pictorial.PictorialOptions>` or
          :obj:`None <python:None>`
        """
        return self._pictorial

    @pictorial.setter
    @class_sensitive(PictorialOptions)
    def pictorial(self, value):
        self._pictorial = value

    @property
    def pie(self) -> Optional[PieOptions]:
        """General options to apply to all Pie series types.

        A pie chart is a circular graphic which is divided into slices to illustrate
        numerical proportion.

        .. tabs::

          .. tab:: Pie Chart

            .. figure:: ../../../_static/pie-example.png
              :alt: Pie Example Chart
              :align: center

          .. tab:: Donut Chart

            .. figure:: ../../../_static/pie-example-donut.png
              :alt: Donut Example Chart
              :align: center

        :rtype: :class:`PieOptions` or :obj:`None <python:None>`
        """
        return self._pie

    @pie.setter
    @class_sensitive(PieOptions)
    def pie(self, value):
        self._pie = value

    @property
    def polygon(self) -> Optional[PolygonOptions]:
        """General options to apply to all Polygon series types.

        A polygon series can be used to draw any freeform shape in the cartesian
        coordinate system. A fill is applied with the :meth:`PolygonOptions.color`
        setting, and stroke is applied through :meth:`PolygonOptions.line_width` and
        :meth:`PolygonOptions.line_color`.

        .. figure:: ../../../_static/polygon-example.png
          :alt: Polygon Example Chart
          :align: center

        :rtype: :class:`PolygonOptions` or :obj:`None <python:None>`
        """
        return self._polygon

    @polygon.setter
    @class_sensitive(PolygonOptions)
    def polygon(self, value):
        self._polygon = value

    @property
    def pyramid(self) -> Optional[PyramidOptions]:
        """General options to apply to all Pyramid series types.

        A pyramid series is a special type of funnel, without neck and reversed by
        default.

        .. figure:: ../../../_static/pyramid-example.png
          :alt: Pyramid Example Chart
          :align: center

        :rtype: :class:`PyramidOptions` or :obj:`None <python:None>`
        """
        return self._pyramid

    @pyramid.setter
    @class_sensitive(PyramidOptions)
    def pyramid(self, value):
        self._pyramid = value

    @property
    def pyramid_3d(self) -> Optional[Pyramid3DOptions]:
        """General options to apply to all Pyramid 3D series types.

        A pyramid 3d series is a special type of funnel, without neck and reversed by
        default.

        .. figure:: ../../../_static/pyramid_3d-example.png
          :alt: Pyramid 3D Example Chart
          :align: center

        :rtype: :class:`Pyramid3DOptions` or :obj:`None <python:None>`
        """
        return self._pyramid_3d

    @pyramid_3d.setter
    @class_sensitive(Pyramid3DOptions)
    def pyramid_3d(self, value):
        self._pyramid_3d = value

    @property
    def sankey(self) -> Optional[SankeyOptions]:
        """General options to apply to all Sankey series types.

        A sankey diagram is a type of flow diagram, in which the width of the link between
        two nodes is shown proportionally to the flow quantity.

        .. tabs::

          .. tab:: Standard Sankey

            .. figure:: ../../../_static/sankey-example.png
              :alt: Sankey Example Chart
              :align: center

          .. tab:: Inverted Sankey

            .. figure:: ../../../_static/sankey-example-inverted.png
              :alt: Inverted Sankey Example Chart
              :align: center

          .. tab:: Sankey with Outgoing Links

            .. figure:: ../../../_static/sankey-example-outgoing.png
              :alt: Sankey Example Chart with Outgoing Links
              :align: center

        :rtype: :class:`SankeyOptions` or :obj:`None <python:None>`
        """
        return self._sankey

    @sankey.setter
    @class_sensitive(SankeyOptions)
    def sankey(self, value):
        self._sankey = value

    @property
    def scatter(self) -> Optional[ScatterOptions]:
        """General options to apply to all Scatter series types.

        A scatter plot uses cartesian coordinates to display values for two variables for
        a set of data.

        .. figure:: ../../../_static/scatter-example.png
          :alt: Scatter Example Chart
          :align: center

        :rtype: :class:`ScatterOptions` or :obj:`None <python:None>`
        """
        return self._scatter

    @scatter.setter
    @class_sensitive(ScatterOptions)
    def scatter(self, value):
        self._scatter = value

    @property
    def scatter_3d(self) -> Optional[Scatter3DOptions]:
        """General options to apply to all Scatter 3D series types.

        A 3D scatter plot uses x, y and z coordinates to display values for three
        variables for a set of data.

        .. figure:: ../../../_static/scatter_3d-example.png
          :alt: Scatter 3D Example Chart
          :align: center

        :rtype: :class:`Scatter3DOptions` or :obj:`None <python:None>`
        """
        return self._scatter_3d

    @scatter_3d.setter
    @class_sensitive(Scatter3DOptions)
    def scatter_3d(self, value):
        self._scatter_3d = value

    @property
    def series(self) -> Optional[SeriesOptions]:
        """General options to apply to all series types.

        :rtype: :class:`ScatterOptions` or :obj:`None <python:None>`
        """
        return self._series

    @series.setter
    @class_sensitive(SeriesOptions)
    def series(self, value):
        self._series = value

    @property
    def solidgauge(self) -> Optional[SolidGaugeOptions]:
        """General options to apply to all Solid Gauge series types.

        A solid gauge is a circular gauge where the value is indicated by a filled arc,
        and the color of the arc may variate with the value.

        .. figure:: ../../../_static/solidgauge-example.png
          :alt: SolidGauge Example Chart
          :align: center

        :rtype: :class:`SolidGaugeOptions` or :obj:`None <python:None>`
        """
        return self._solidgauge

    @solidgauge.setter
    @class_sensitive(SolidGaugeOptions)
    def solidgauge(self, value):
        self._solidgauge = value

    @property
    def spline(self) -> Optional[SplineOptions]:
        """General options to apply to all Spline series types.

        A spline series is a special type of line series, where the segments between the
        data points are smoothed.

        .. figure:: ../../../_static/spline-example.png
          :alt: Spline Example Chart
          :align: center

        :rtype: :class:`SplineOptions` or :obj:`None <python:None>`
        """
        return self._spline

    @spline.setter
    @class_sensitive(SplineOptions)
    def spline(self, value):
        self._spline = value

    @property
    def streamgraph(self) -> Optional[StreamGraphOptions]:
        """General options to apply to all Stream Graph series types.

        A streamgraph is a type of stacked area graph which is displaced around a central
        axis, resulting in a flowing, organic shape.

        .. figure:: ../../../_static/streamgraph-example.png
          :alt: StreamGraph Example Chart
          :align: center

        :rtype: :class:`StreamGraphOptions` or :obj:`None <python:None>`
        """
        return self._streamgraph

    @streamgraph.setter
    @class_sensitive(StreamGraphOptions)
    def streamgraph(self, value):
        self._streamgraph = value

    @property
    def sunburst(self) -> Optional[SunburstOptions]:
        """General options to apply to all Sunburst series types.

        A Sunburst displays hierarchical data, where a level in the hierarchy is
        represented by a circle. The center represents the root node of the tree. The
        visualization bears a resemblance to both treemap and pie charts.

        .. figure:: ../../../_static/sunburst-example.png
          :alt: Sunburst Example Chart
          :align: center

        :rtype: :class:`SunburstOptions` or :obj:`None <python:None>`
        """
        return self._sunburst

    @sunburst.setter
    @class_sensitive(SunburstOptions)
    def sunburst(self, value):
        self._sunburst = value

    @property
    def tilemap(self) -> Optional[TilemapOptions]:
        """General options to apply to all Tilemap series types.

        A tilemap series is a type of heatmap where the tile shapes are configurable.

        .. tabs::

          .. tab:: Honeycomb Tilemap

            .. figure:: ../../../_static/tilemap-example.png
              :alt: Honeycomb Tilemap Example Chart
              :align: center

          .. tab:: Circle Tilemap

            .. figure:: ../../../_static/tilemap-example-circle.png
              :alt: Tilemap Example Chart
              :align: center

          .. tab:: Diamond Tilemap

            .. figure:: ../../../_static/tilemap-example-diamond.png
              :alt: Tilemap Example Chart
              :align: center

        :rtype: :class:`TilemapOptions` or :obj:`None <python:None>`
        """
        return self._tilemap

    @tilemap.setter
    @class_sensitive(TilemapOptions)
    def tilemap(self, value):
        self._tilemap = value

    @property
    def timeline(self) -> Optional[TimelineOptions]:
        """General options to apply to all Timeline series types.

        The timeline series presents given events along a drawn line.

        .. tabs::

          .. tab:: Standard Timeline

            .. figure:: ../../../_static/timeline-example.png
              :alt: Timeline Example Chart
              :align: center

          .. tab:: Inverted Timeline

            .. figure:: ../../../_static/timeline-example-inverted.png
              :alt: Inverted Timeline Example Chart
              :align: center

          .. tab:: With True Datetime Axis

            .. figure:: ../../../_static/timeline-example-datetime.png
              :alt: Timeline Example Chart with Datetime Axis
              :align: center

        :rtype: :class:`TimelineOptions` or :obj:`None <python:None>`
        """
        return self._timeline

    @timeline.setter
    @class_sensitive(TimelineOptions)
    def timeline(self, value):
        self._timeline = value

    @property
    def treegraph(self) -> Optional[TreegraphOptions]:
        """General options to apply to all :term:`Treegraph` series types.
        
        A treegraph visualizes a relationship between ancestors and descendants with a clear parent-child relationship,
        e.g. a family tree or a directory structure.
        
        .. figure:: ../../../_static/treegraph-example.png
          :alt: Treegraph Example Chart
          :align: center
        
        :rtype: :class:`TreegraphOptions <highcharts_core.options.plot_options.treegraph.TreegraphOptions>` or 
          :obj:`None <python:None>`
        """
        return self._treegraph
    
    @treegraph.setter
    @class_sensitive(TreegraphOptions)
    def treegraph(self, value):
        self._treegraph = value

    @property
    def treemap(self) -> Optional[TreemapOptions]:
        """General options to apply to all Treemap series types.

        A treemap displays hierarchical data using nested rectangles. The data can be laid
        out in varying ways depending on options.

        .. figure:: ../../../_static/treemap-example.png
          :alt: Treemap Example Chart
          :align: center

        :rtype: :class:`TreemapOptions` or :obj:`None <python:None>`
        """
        return self._treemap

    @treemap.setter
    @class_sensitive(TreemapOptions)
    def treemap(self, value):
        self._treemap = value

    @property
    def variablepie(self) -> Optional[VariablePieOptions]:
        """General options to apply to all Variable Pie series types.

        A variable pie series is a two dimensional series type, where each point renders
        an Y and Z value. Each point is drawn as a pie slice where the size (arc) of the
        slice relates to the Y value and the radius of pie slice relates to the Z value.

        .. figure:: ../../../_static/variablepie-example.png
          :alt: Variable Pie Example Chart
          :align: center

        :rtype: :class:`VariablePieOptions` or :obj:`None <python:None>`
        """
        return self._variablepie

    @variablepie.setter
    @class_sensitive(VariablePieOptions)
    def variablepie(self, value):
        self._variablepie = value

    @property
    def variwide(self) -> Optional[VariwideOptions]:
        """General options to apply to all Variwide series types.

        A variwide chart (related to marimekko chart) is a column chart with a variable
        width expressing a third dimension.

        .. tabs::

          .. tab:: Standard Variwide

            .. figure:: ../../../_static/variwide-example.png
              :alt: Variwide Example Chart
              :align: center

          .. tab:: Inverted Variwide

            .. figure:: ../../../_static/variwide-example-inverted.png
              :alt: Variwide Example Chart
              :align: center

          .. tab:: with Datetime Axis

            .. figure:: ../../../_static/variwide-example-datetime.png
              :alt: Variwide Example Chart
              :align: center

        :rtype: :class:`VariwideOptions` or :obj:`None <python:None>`
        """
        return self._variwide

    @variwide.setter
    @class_sensitive(VariwideOptions)
    def variwide(self, value):
        self._variwide = value

    @property
    def vector(self) -> Optional[VectorOptions]:
        """General options to apply to all Vector series types.

        A vector plot is a type of cartesian chart where each point has an X and Y
        position, a length and a direction. Vectors are drawn as arrows.

        .. figure:: ../../../_static/vector-example.png
          :alt: Vector Example Chart
          :align: center

        :rtype: :class:`VectorOptions` or :obj:`None <python:None>`
        """
        return self._vector

    @vector.setter
    @class_sensitive(VectorOptions)
    def vector(self, value):
        self._vector = value

    @property
    def venn(self) -> Optional[VennOptions]:
        """General options to apply to all Venn series types.

        A Venn diagram displays all possible logical relations between a collection of
        different sets. The sets are represented by circles, and the relation between the
        sets are displayed by the overlap or lack of overlap between them. The venn
        diagram is a special case of Euler diagrams, which can also be displayed by this
        series type.

        .. tabs::

          .. tab:: Venn Diagram

            .. figure:: ../../../_static/venn-example.png
              :alt: Venn Example Chart
              :align: center

          .. tab:: Euler Diagram

            .. figure:: ../../../_static/venn-example-euler.png
              :alt: Euler Example Chart
              :align: center

        :rtype: :class:`VennOptions` or :obj:`None <python:None>`
        """
        return self._venn

    @venn.setter
    @class_sensitive(VennOptions)
    def venn(self, value):
        self._venn = value

    @property
    def waterfall(self) -> Optional[WaterfallOptions]:
        """General options to apply to all Waterfall series types.

        A waterfall chart displays sequentially introduced positive or negative values in
        cumulative columns.

        .. tabs::

          .. tab:: Standard Waterfall

            .. figure:: ../../../_static/waterfall-example.png
              :alt: Waterfall Example Chart
              :align: center

          .. tab:: Horizontal (Inverted) Waterfall

            .. figure:: ../../../_static/waterfall-example-inverted.png
              :alt: Waterfall Example Chart
              :align: center

          .. tab:: Stacked Waterfall

            .. figure:: ../../../_static/waterfall-example-stacked.png
              :alt: Waterfall Example Chart
              :align: center

        :rtype: :class:`WaterfallOptions` or :obj:`None <python:None>`
        """
        return self._waterfall

    @waterfall.setter
    @class_sensitive(WaterfallOptions)
    def waterfall(self, value):
        self._waterfall = value

    @property
    def windbarb(self) -> Optional[WindBarbOptions]:
        """General options to apply to all Wind Barb series types.

        Wind barbs are a convenient way to represent wind speed and direction in one
        graphical form. Wind direction is given by the stem direction, and wind speed by
        the number and shape of barbs.

        .. figure:: ../../../_static/windbarb-example.png
          :alt: Wind Barb Example Chart
          :align: center

        :rtype: :class:`WindBarbOptions` or :obj:`None <python:None>`
        """
        return self._windbarb

    @windbarb.setter
    @class_sensitive(WindBarbOptions)
    def windbarb(self, value):
        self._windbarb = value

    @property
    def wordcloud(self) -> Optional[WordcloudOptions]:
        """General options to apply to all Wordcloud series types.

        A word cloud is a visualization of a set of words, where the size and placement of
        a word is determined by how it is weighted.

        .. figure:: ../../../_static/wordcloud-example.png
          :alt: Wordcloud Example Chart
          :align: center

        :rtype: :class:`WordcloudOptions` or :obj:`None <python:None>`
        """
        return self._wordcloud

    @wordcloud.setter
    @class_sensitive(WordcloudOptions)
    def wordcloud(self, value):
        self._wordcloud = value

    @property
    def xrange(self) -> Optional[XRangeOptions]:
        """General options to apply to all X-Range series types.

        The X-range series displays ranges on the X axis, typically time intervals with a
        start and end date.

        .. tabs::

          .. tab:: Standard X-Range

            .. figure:: ../../../_static/xrange-example.png
              :alt: X-Range Example Chart
              :align: center

          .. tab:: Inverted X-Range

            .. figure:: ../../../_static/xrange-example-inverted.png
              :alt: Inverted X-Range Example Chart
              :align: center

        :rtype: :class:`XRangeOptions` or :obj:`None <python:None>`
        """
        return self._xrange

    @xrange.setter
    @class_sensitive(XRangeOptions)
    def xrange(self, value):
        self._xrange = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'arcdiagram': as_dict.get('arcdiagram', None),
            'area': as_dict.get('area', None),
            'arearange': as_dict.get('arearange', None),
            'areaspline': as_dict.get('areaspline', None),
            'areasplinerange': as_dict.get('areasplinerange', None),
            'bar': as_dict.get('bar', None),
            'bellcurve': as_dict.get('bellcurve', None),
            'boxplot': as_dict.get('boxplot', None),
            'bubble': as_dict.get('bubble', None),
            'bullet': as_dict.get('bullet', None),
            'column': as_dict.get('column', None),
            'columnpyramid': as_dict.get('columnpyramid', None),
            'columnrange': as_dict.get('columnrange', None),
            'cylinder': as_dict.get('cylinder', None),
            'dependencywheel': as_dict.get('dependencywheel', None),
            'dumbbell': as_dict.get('dumbbell', None),
            'errorbar': as_dict.get('errorbar', None),
            'funnel': as_dict.get('funnel', None),
            'funnel_3d': as_dict.get('funnel3d', None),
            'gauge': as_dict.get('gauge', None),
            'heatmap': as_dict.get('heatmap', None),
            'histogram': as_dict.get('histogram', None),
            'item': as_dict.get('item', None),
            'line': as_dict.get('line', None),
            'lollipop': as_dict.get('lollipop', None),
            'networkgraph': as_dict.get('networkgraph', None),
            'organization': as_dict.get('organization', None),
            'packedbubble': as_dict.get('packedbubble', None),
            'pareto': as_dict.get('pareto', None),
            'pictorial': as_dict.get('pictorial', None),
            'pie': as_dict.get('pie', None),
            'polygon': as_dict.get('polygon', None),
            'pyramid': as_dict.get('pyramid', None),
            'pyramid_3d': as_dict.get('pyramid3d', None),
            'sankey': as_dict.get('sankey', None),
            'scatter': as_dict.get('scatter', None),
            'scatter_3d': as_dict.get('scatter3d', None),
            'series': as_dict.get('series', None),
            'solidgauge': as_dict.get('solidgauge', None),
            'spline': as_dict.get('spline', None),
            'streamgraph': as_dict.get('streamgraph', None),
            'sunburst': as_dict.get('sunburst', None),
            'tilemap': as_dict.get('tilemap', None),
            'timeline': as_dict.get('timeline', None),
            'treegraph': as_dict.get('treegraph', None),
            'treemap': as_dict.get('treemap', None),
            'variablepie': as_dict.get('variablepie', None),
            'variwide': as_dict.get('variwide', None),
            'vector': as_dict.get('vector', None),
            'venn': as_dict.get('venn', None),
            'waterfall': as_dict.get('waterfall', None),
            'windbarb': as_dict.get('windbarb', None),
            'wordcloud': as_dict.get('wordcloud', None),
            'xrange': as_dict.get('xrange', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'arcdiagram': self.arcdiagram,
            'area': self.area,
            'arearange': self.arearange,
            'areaspline': self.areaspline,
            'areasplinerange': self.areasplinerange,
            'bar': self.bar,
            'bellcurve': self.bellcurve,
            'boxplot': self.boxplot,
            'bubble': self.bubble,
            'bullet': self.bullet,
            'column': self.column,
            'columnpyramid': self.columnpyramid,
            'columnrange': self.columnrange,
            'cylinder': self.cylinder,
            'dependencywheel': self.dependencywheel,
            'dumbbell': self.dumbbell,
            'errorbar': self.errorbar,
            'funnel': self.funnel,
            'funnel3d': self.funnel_3d,
            'gauge': self.gauge,
            'heatmap': self.heatmap,
            'histogram': self.histogram,
            'item': self.item,
            'line': self.line,
            'lollipop': self.lollipop,
            'networkgraph': self.networkgraph,
            'organization': self.organization,
            'packedbubble': self.packedbubble,
            'pareto': self.pareto,
            'pictorial': self.pictorial,
            'pie': self.pie,
            'polygon': self.polygon,
            'pyramid': self.pyramid,
            'pyramid3d': self.pyramid_3d,
            'sankey': self.sankey,
            'scatter': self.scatter,
            'scatter3d': self.scatter_3d,
            'series': self.series,
            'solidgauge': self.solidgauge,
            'spline': self.spline,
            'streamgraph': self.streamgraph,
            'sunburst': self.sunburst,
            'tilemap': self.tilemap,
            'timeline': self.timeline,
            'treegraph': self.treegraph,
            'treemap': self.treemap,
            'variablepie': self.variablepie,
            'variwide': self.variwide,
            'vector': self.vector,
            'venn': self.venn,
            'waterfall': self.waterfall,
            'windbarb': self.windbarb,
            'wordcloud': self.wordcloud,
            'xrange': self.xrange
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.gradients import Gradient as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'linear_gradient': {
          'x1': 0.123,
          'x2': 0.567,
          'y1': 0.891,
          'y2': 0.987
      },
      'stops': [
          [0.123, '#cccccc'],
          [0.456, '#ff0000'],
          [1, '#00ff00']
      ]
    }, None),
    ({
     'radial_gradient': {
         'cx': 0.123,
         'cy': 0.456,
         'r': 0.789
     },
      'stops': [
          [0.123, '#cccccc'],
          [0.456, '#ff0000'],
          [1, '#00ff00']
      ]
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/gradients/01.js', False, None),
    ('utility_classes/gradients/02.js', False, None),

    ('utility_classes/gradients/error-01.js', False, (errors.HighchartsValueError,
                                                      errors.HighchartsParseError,
                                                      JSONDecodeError,
                                                      TypeError,
                                                      ValueError)),
    ('utility_classes/gradients/error-02.js', False, (errors.HighchartsValueError,
                                                      errors.HighchartsParseError,
                                                      JSONDecodeError,
                                                      TypeError,
                                                      ValueError)),
    ('utility_classes/gradients/error-03.js', True, (errors.HighchartsValueError,
                                                     errors.HighchartsParseError,
                                                     JSONDecodeError,
                                                     TypeError,
                                                     ValueError)),

    ('utility_classes/gradients/01.js', True, None),
    ('utility_classes/gradients/02.js', True, None),

    ('utility_classes/gradients/error-01.js', True, (errors.HighchartsValueError,
                                                     errors.HighchartsParseError,
                                                     JSONDecodeError,
                                                     TypeError,
                                                     ValueError)),
    ('utility_classes/gradients/error-02.js', True, (errors.HighchartsValueError,
                                                     errors.HighchartsParseError,
                                                     JSONDecodeError,
                                                     TypeError,
                                                     ValueError)),
    ('utility_classes/gradients/error-03.js', True, (errors.HighchartsValueError,
                                                     errors.HighchartsParseError,
                                                     JSONDecodeError,
                                                     TypeError,
                                                     ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



"""Tests for ``highcharts.no_data``."""
from copy import deepcopy
import pytest

from json.decoder import JSONDecodeError
from validator_collection import checkers

from highcharts_core.headless_export import ExportServer as cls
from highcharts_core.options import HighchartsOptions
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal, run_download_tests, create_output_directory

STANDARD_PARAMS = [
    ({}, None),
]

@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)

@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    from tests.fixtures import trim_expected, does_kwarg_value_match_result
    from validator_collection import checkers

    kwargs_copy = deepcopy(kwargs)
    if not kwargs:
        kwargs_copy['url'] = 'https://export.highcharts.com'
        kwargs_copy['format'] = 'png'
        kwargs_copy['scale'] = 1
        kwargs_copy['use_base64'] = False
        kwargs_copy['no_download'] = False
        kwargs_copy['async_rendering'] = False
        kwargs_copy['constructor'] = 'Chart'

        untrimmed_expected = {
            'url': kwargs_copy['url'],
            'type': kwargs_copy['format'],
            'scale': kwargs_copy['scale'],
            'constr': kwargs_copy['constructor'],
            'b64': kwargs_copy['use_base64'],
            'noDownload': kwargs_copy['no_download'],
            'asyncRendering': kwargs_copy['async_rendering']
        }

    else:
        untrimmed_expected = to_js_dict(deepcopy(kwargs))

    expected = trim_expected(untrimmed_expected)
    check_dicts = True
    for key in expected:
        if not checkers.is_type(expected[key], (str, int, float, bool, list, dict)):
            check_dicts = False
        elif isinstance(expected[key], str) and expected[key].startswith('function'):
            check_dicts = False
        elif isinstance(expected[key], str) and expected[key].startswith('class'):
            check_dicts = False

    if not error:
        instance = cls(**kwargs)
        result = instance.to_dict()
        assert result is not None
        assert isinstance(result, dict) is True
        print('EXPECTED:')
        print(expected)
        print('RESULT:')
        print(result)
        if check_dicts:
            assert len(expected) == len(result)
            for key in expected:
                print(f'CHECKING: {key}')
                assert does_kwarg_value_match_result(expected[key],
                                                     result.get(key)) is True
    else:
        with pytest.raises(error):
            instance = cls(**kwargs)
            result = instance.to_dict()


@pytest.mark.parametrize('value, error', [
    ({
        'url': 'https://export.highcharts.com',
        'protocol': 'https',
        'domain': 'export.highcharts.com',
        'port': None,
        'path': None
    }, None),
    ({
        'url': 'https://export.highcharts.com:1234/some_path_goes_here/andcontinueshere',
        'protocol': 'https',
        'domain': 'export.highcharts.com',
        'port': 1234,
        'path': 'some_path_goes_here/andcontinueshere'
    }, None),
    ({
        'url': 'http://www.highcharts.com/some_path_goes_here/andcontinueshere',
        'protocol': 'http',
        'domain': 'www.highcharts.com',
        'port': None,
        'path': 'some_path_goes_here/andcontinueshere'
    }, None),
    ({
        'url': 'http://www.highcharts.com:8080',
        'protocol': 'http',
        'domain': 'www.highcharts.com',
        'port': 8080,
        'path': None
    }, None),
])
def test_url(value, error):
    instance = cls()
    if not error:
        if 'expected_url' not in value and 'url' in value:
            value['expected_url'] = value.get('url')

        url = value.get('url', None)

        instance.url = url
        assert instance.url is not None
        assert instance.url == value.get('expected_url')
        assert instance.protocol == value.get('protocol')
        assert instance.domain == value.get('domain')
        assert instance.port == value.get('port')
        assert instance.path == value.get('path')
    else:
        with pytest.raises(error):
            instance.url = value.get('url')


@pytest.mark.parametrize('input_filename, target_filename, kwargs, error', [
    ('headless_export/basic.json',
     'headless_export/output/test-basic.png',
     {
      'timeout': 5,
      'format_': 'png',
      'constructor': 'Chart'
     },
     None),
    ('headless_export/with-series-types.js',
     'headless_export/output/test-with-series-types.png',
     {
      'timeout': 5,
      'format_': 'png',
      'constructor': 'Chart'
     },
     None),
    ('headless_export/with-chart-type.js',
     'headless_export/output/test-with-chart-type.png',
     {
      'timeout': 5,
      'format_': 'png',
      'constructor': 'Chart'
     },
     None),
    ('headless_export/with-chart-type.js',
     'headless_export/output/test-with-chart-type.svg',
     {
      'timeout': 5,
      'format_': 'svg',
      'constructor': 'Chart'
     },
     None),
])
def test_get_chart(input_files,
                   run_download_tests,
                   create_output_directory,
                   input_filename,
                   target_filename,
                   kwargs,
                   error):
    if run_download_tests:
        input_file = check_input_file(input_files, input_filename)
        target_file = check_input_file(input_files, 
                                       target_filename,
                                       create_directory = create_output_directory)
        
        with open(input_file, 'r') as file_:
            as_str = file_.read()

        options = HighchartsOptions.from_js_literal(as_str)

        kwargs['options'] = options

        if target_filename:
            kwargs['filename'] = target_file

        if not error:
            result = cls.get_chart(**kwargs)
    
            format = kwargs.get('format_', None)
            print(f'TESTING FORMAT: {format}')
    
            assert result is not None
            if target_filename:
                assert checkers.is_on_filesystem(target_file) is True
                if format == 'svg':
                    with open(target_file, 'r', encoding = 'utf-8') as file_:
                        file_contents = file_.read()
                    contents = str(file_contents)
                    assert contents.startswith(
                        '<?xml version="1.0" standalone="no"?><!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"'
                    ) is True



from typing import Optional, List, Any
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.options.plot_options.series import SeriesOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.shadows import ShadowOptions


class DialOptions(HighchartsMeta):
    """Options for the dial or arrow pointer of the gauge."""

    def __init__(self, **kwargs):
        self._background_color = None
        self._base_length = None
        self._base_width = None
        self._border_color = None
        self._border_width = None
        self._path = None
        self._radius = None
        self._rear_length = None
        self._top_width = None

        self.background_color = kwargs.get('background_color', None)
        self.base_length = kwargs.get('base_length', None)
        self.base_width = kwargs.get('base_width', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.path = kwargs.get('path', None)
        self.radius = kwargs.get('radius', None)
        self.rear_length = kwargs.get('rear_length', None)
        self.top_width = kwargs.get('top_width', None)

    @property
    def background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background or fill color of the gauge's dial. Defaults to '#000000'.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        from highcharts_core import utility_functions
        self._background_color = utility_functions.validate_color(value)

    @property
    def base_length(self) -> Optional[str]:
        """The length of the dial's base part, expressed as a percentage of the total
        radius or length of the dial. Defaults to ``70%``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._base_length

    @base_length.setter
    def base_length(self, value):
        if value is None:
            self._base_length = None
        else:
            value = validators.string(value)
            if '%' not in value:
                raise errors.HighchartsValueError(f'base_length expects a percentage '
                                                  f'string. Received: {value}')
            self._base_length = value

    @property
    def base_width(self) -> Optional[int | float | Decimal]:
        """The pixel width of the base of the gauge dial. Defaults to ``3``.

        .. note::

          The base is the part closest to the pivot, defined by
          :meth:`DialOptions.base_length`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._base_width

    @base_width.setter
    def base_width(self, value):
        self._base_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The border color or stroke of the gauge's dial. Defaults to ``'#cccccc'``.

        .. warning::

          By default, the :meth:`DialOptions.border_width` is ``0``, so be sure to set
          that property to a non-default value if you wish to apply a border color.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The width of the gauge dial border in pixels. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def path(self) -> Optional[List[Any]]:
        """An array with an SVG path for the custom dial. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`list <python:list>` of any instances, or :obj:`None <python:None>`
        """
        return self._path

    @path.setter
    def path(self, value):
        if not value:
            self._path = None
        else:
            self._path = validators.iterable(value)

    @property
    def radius(self) -> Optional[str]:
        """The radius or length of the dial, in percentages relative to the radius of the
        gauge itself. Defaults to ``'80%'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        if not value:
            self._radius = None
        else:
            value = validators.string(value)
            if '%' not in value:
                raise errors.HighchartsValueError(f'radius expects a percentage string. '
                                                  f'Received: {value}')
            self._radius = value

    @property
    def rear_length(self) -> Optional[str]:
        """The length of the dial's rear end, the part that extends out on the other side
        of the pivot. Expressed as a percentage of the dial's length. Defaults to
        ``'10%'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._rear_length

    @rear_length.setter
    def rear_length(self, value):
        if not value:
            self._rear_length = None
        else:
            value = validators.string(value)
            if '%' not in value:
                raise errors.HighchartsValueError(f'rear_length expects a percentage '
                                                  f'string. Received: {value}')
            self._rear_length = value

    @property
    def top_width(self) -> Optional[int | float | Decimal]:
        """The width of the top of the dial, closest to the perimeter. Defaults to ``1``.

        .. note::

          The pivot narrows in from the base to the top.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._top_width

    @top_width.setter
    def top_width(self, value):
        self._top_width = validators.numeric(value, minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'background_color': as_dict.get('backgroundColor', None),
            'base_length': as_dict.get('baseLength', None),
            'base_width': as_dict.get('baseWidth', None),
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'path': as_dict.get('path', None),
            'radius': as_dict.get('radius', None),
            'rear_length': as_dict.get('rearLength', None),
            'top_width': as_dict.get('topWidth', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'backgroundColor': self.background_color,
            'baseLength': self.base_length,
            'baseWidth': self.base_width,
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'path': self.path,
            'radius': self.radius,
            'rearLength': self.rear_length,
            'topWidth': self.top_width
        }

        return untrimmed


class PivotOptions(HighchartsMeta):
    """Options for the pivot or the center point of the gauge."""

    def __init__(self, **kwargs):
        self._background_color = None
        self._border_color = None
        self._border_width = None
        self._radius = None

        self.background_color = kwargs.get('background_color', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.radius = kwargs.get('radius', None)

    @property
    def background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or fill of the pivot. Defaults to '#000000'.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        from highcharts_core import utility_functions
        self._background_color = utility_functions.validate_color(value)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The border or stroke color of the pivot. Defaults to ``'#cccccc'``.

        .. warning::

          By default, the :meth:`PivotOptions.border_width` is ``0``, so be sure to set
          that property to a non-default value if you wish to apply a border color.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """TThe border or stroke width of the pivot. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def radius(self) -> Optional[int | float | Decimal]:
        """The pixel radius of the pivot. Defaults to ``5``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'radius': as_dict.get('radius', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'backgroundColor': self.background_color,
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'radius': self.radius
        }

        return untrimmed


class GaugeOptions(GenericTypeOptions):
    """General options to apply to all Gauge series types.

    Gauges are circular plots displaying one or more values with a dial pointing to
    values along the perimeter.

    .. figure:: ../../../_static/gauge-example.png
      :alt: Gauge Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._color_index = None
        self._crisp = None
        self._dial = None
        self._linecap = None
        self._line_width = None
        self._overshoot = None
        self._pivot = None
        self._point_interval = None
        self._point_interval_unit = None
        self._point_start = None
        self._relative_x_value = None
        self._shadow = None
        self._wrap = None

        self.color_index = kwargs.get('color_index', None)
        self.crisp = kwargs.get('crisp', None)
        self.dial = kwargs.get('dial', None)
        self.linecap = kwargs.get('linecap', None)
        self.line_width = kwargs.get('line_width', None)
        self.overshoot = kwargs.get('overshoot', None)
        self.pivot = kwargs.get('pivot', None)
        self.point_interval = kwargs.get('point_interval', None)
        self.point_interval_unit = kwargs.get('point_interval_unit', None)
        self.point_start = kwargs.get('point_start', None)
        self.relative_x_value = kwargs.get('relative_x_value', None)
        self.shadow = kwargs.get('shadow', None)
        self.wrap = kwargs.get('wrap', None)

        super().__init__(**kwargs)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        series, so that its graphic representations are given the class name
        ``highcharts-color-{n}``.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def crisp(self) -> Optional[bool]:
        """If ``True``, each point or column edge is rounded to its nearest pixel in order
        to render sharp on screen. Defaults to ``True``.

        .. hint::

          In some cases, when there are a lot of densely packed columns, this leads to
          visible difference in column widths or distance between columns. In these cases,
          setting ``crisp`` to ``False`` may look better, even though each column is
          rendered blurry.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crisp

    @crisp.setter
    def crisp(self, value):
        if value is None:
            self._crisp = None
        else:
            self._crisp = bool(value)

    @property
    def dial(self) -> Optional[DialOptions]:
        """Options for the dial or arrow pointer of the gauge.

        :rtype: :class:`DialOptions` or :obj:`None <python:None>`
        """
        return self._dial

    @dial.setter
    @class_sensitive(DialOptions)
    def dial(self, value):
        self._dial = value

    @property
    def linecap(self) -> Optional[str]:
        """The SVG value used for the ``stroke-linecap`` and ``stroke-linejoin`` of a line
        graph. Defaults to ``'round'``, which means that lines are rounded in the ends and
        bends.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._linecap

    @linecap.setter
    def linecap(self, value):
        self._linecap = validators.string(value, allow_empty = True)

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """Pixel width of the graph line. Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def overshoot(self) -> Optional[int | float | Decimal]:
        """Allow the dial to overshoot the end of the perimeter axis by this many degrees.

        For example, if this option is set to ``5`` and the gauge axis goes from ``0`` to
        ``60``, a value of ``100``, or ``1000``, will show ``5`` degrees beyond the end of
        the axis.

        Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._overshoot

    @overshoot.setter
    def overshoot(self, value):
        self._overshoot = validators.numeric(value,
                                             allow_empty = True,
                                             minimum = 0)

    @property
    def pivot(self) -> Optional[PivotOptions]:
        """Options for the pivot or the center point of the gauge.

        :rtype: :class:`PivotOptions` or :obj:`None <python:None>`
        """
        return self._pivot

    @pivot.setter
    @class_sensitive(PivotOptions)
    def pivot(self, value):
        self._pivot = value

    @property
    def point_interval(self) -> Optional[int | float | Decimal]:
        """If no x values are given for the points in a series, ``point_interval`` defines
        the interval of the x values. Defaults to ``1``.

        For example, if a series contains one value every decade starting from year 0, set
        ``point_interval`` to ``10``. In true datetime axes, the ``point_interval`` is set
        in milliseconds.

        .. hint::

          ``point_interval`` can be also be combined with
          :meth:`point_interval_unit <AreaOptions.point_interval_unit>` to draw irregular
          time intervals.

        .. note::

          If combined with :meth:`relative_x_value <AreaOptions.relative_x_value>`, an x
          value can be set on each point, and the ``point_interval`` is added x times to
          the :meth:`point_start <AreaOptions.point_start>` setting.

        .. warning::

          This options applies to the series data, not the interval of the axis ticks,
          which is independent.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_interval

    @point_interval.setter
    def point_interval(self, value):
        self._point_interval = validators.numeric(value,
                                                  allow_empty = True,
                                                  minimum = 0)

    @property
    def point_interval_unit(self) -> Optional[str]:
        """On datetime series, this allows for setting the
        :meth:`point_interval <AreaOptions.point_interval>` to irregular time units, day,
        month, and year.

        A day is usually the same as 24 hours, but ``point_interval_unit`` also takes the
        DST crossover into consideration when dealing with local time.

        Combine this option with :meth:`point_interval <AreaOptions.point_interval>` to
        draw weeks, quarters, 6 month periods, 10 year periods, etc.

        .. warning::

          This options applies to the series data, not the interval of the axis ticks,
          which is independent.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._point_interval_unit

    @point_interval_unit.setter
    def point_interval_unit(self, value):
        self._point_interval_unit = validators.string(value, allow_empty = True)

    @property
    def point_start(self) -> Optional[int | float | Decimal]:
        """If no x values are given for the points in a series, ``point_start`` defines
        on what value to start. For example, if a series contains one yearly value
        starting from 1945, set ``point_start`` to ``1945``. Defaults to ``0``.

        .. note::

          If combined with :meth:`relative_x_value <AreaOptions.relative_x_value>`, an x
          value can be set on each point. The x value from the point options is multiplied
          by :meth:`point_interval <AreaOptions.point_interval>` and added to
          ``point_start`` to produce a modified x value.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_start

    @point_start.setter
    def point_start(self, value):
        self._point_start = validators.numeric(value, allow_empty = True)

    @property
    def relative_x_value(self) -> Optional[bool]:
        """When ``True``, X values in the data set are relative to the current
        :meth:`point_start <AreaOptions.point_start>`,
        :meth:`point_interval <AreaOptions.point_interval>`, and
        :meth:`point_interval_unit <AreaOptions.point_interval_unit>` settings. This
        allows compression of the data for datasets with irregular X values. Defaults to
        ``False``.

        The real X values are computed on the formula ``f(x) = ax + b``, where ``a`` is
        the :meth:`point_interval <AreaOptions.point_interval>` (optionally with a time
        unit given by :meth:`point_interval_unit <AreaOptions.point_interval_unit>`), and
        ``b`` is the :meth:`point_start <AreaOptions.point_start>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._relative_x_value

    @relative_x_value.setter
    def relative_x_value(self, value):
        if value is None:
            self._relative_x_value = None
        else:
            self._relative_x_value = bool(value)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration for the shadow to apply to the tooltip. Defaults to
        ``False``.

        If ``False``, no shadow is applied.

        :returns: The shadow configuration to apply or a boolean setting which hides the
          shadow or displays the default shadow.
        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if isinstance(value, bool):
            self._shadow = value
        elif not value:
            self._shadow = None
        else:
            value = validate_types(value,
                                   types = ShadowOptions)
            self._shadow = value

    @property
    def wrap(self) -> Optional[bool]:
        """When ``True``, the dial will wrap around the axes. When wrap is ``False``, the
        dial stops at 360. Defaults to ``True``.

        For instance, if ``True`` in a full-range gauge going from ``0`` to ``360``, a
        value of ``400`` will point to ``40``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._wrap

    @wrap.setter
    def wrap(self, value):
        if value is None:
            self._wrap = None
        else:
            self._wrap = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'color_index': as_dict.get('colorIndex', None),
            'crisp': as_dict.get('crisp', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),

            'dial': as_dict.get('dial', None),
            'overshoot': as_dict.get('overshoot', None),
            'pivot': as_dict.get('pivot', None),
            'wrap': as_dict.get('wrap', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'colorIndex': self.color_index,
            'crisp': self.crisp,
            'dial': self.dial,
            'linecap': self.linecap,
            'lineWidth': self.line_width,
            'overshoot': self.overshoot,
            'pivot': self.pivot,
            'pointInterval': self.point_interval,
            'pointIntervalUnit': self.point_interval_unit,
            'pointStart': self.point_start,
            'relativeXValue': self.relative_x_value,
            'shadow': self.shadow,
            'wrap': self.wrap
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class SolidGaugeOptions(SeriesOptions):
    """General options to apply to all Solid Gauge series types.

    A solid gauge is a circular gauge where the value is indicated by a filled arc,
    and the color of the arc may variate with the value.

    .. figure:: ../../../_static/solidgauge-example.png
      :alt: SolidGauge Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._inner_radius = None
        self._overshoot = None
        self._radius = None
        self._rounded = None

        self.inner_radius = kwargs.get('inner_radius', None)
        self.overshoot = kwargs.get('overshoot', None)
        self.radius = kwargs.get('radius', None)
        self.rounded = kwargs.get('rounded', None)

        super().__init__(**kwargs)

    @property
    def inner_radius(self) -> Optional[str | int | float | Decimal]:
        """The inner radius for points in a solid gauge. Can be given as a number (pixels)
        or percentage string. Defaults to ``60``.

        :rtype: numeric, :class:`str <python:str>`, or :obj:`None <python:None>`
        """
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value):
        if value is None:
            self._inner_radius = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('inner_radius expects either a '
                                                      'number or a percentage % string. '
                                                      'No "%" character found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._inner_radius = value

    @property
    def overshoot(self) -> Optional[int | float | Decimal]:
        """Allow the dial to overshoot the end of the perimeter axis by this many degrees.

        For example, if this option is set to ``5`` and the gauge axis goes from ``0`` to
        ``60``, a value of ``100``, or ``1000``, will show ``5`` degrees beyond the end of
        the axis.

        Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._overshoot

    @overshoot.setter
    def overshoot(self, value):
        self._overshoot = validators.numeric(value,
                                             allow_empty = True,
                                             minimum = 0)

    @property
    def radius(self) -> Optional[str | int | float | Decimal]:
        """The outer radius for points in a solid gauge. Can be given as a number (pixels)
        or percentage string. Defaults to ``100``.

        :rtype: numeric, :class:`str <python:str>`, or :obj:`None <python:None>`
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        if value is None:
            self._radius = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('radius expects either a number or '
                                                      'a percentage string. No "%" '
                                                      'character found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._radius = value

    @property
    def rounded(self) -> Optional[bool]:
        """If ``True``, draws rounded edges on the gauge. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._rounded

    @rounded.setter
    def rounded(self, value):
        if value is None:
            self._rounded = None
        else:
            self._rounded = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'inner_radius': as_dict.get('innerRadius', None),
            'overshoot': as_dict.get('overshoot', None),
            'radius': as_dict.get('radius', None),
            'rounded': as_dict.get('rounded', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'innerRadius': self.inner_radius,
            'overshoot': self.overshoot,
            'radius': self.radius,
            'rounded': self.rounded
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional, List

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.accessibility.keyboard_navigation.focus_border import (FocusBorder,
                                                                                      FocusBorderStyle)
from highcharts_core.options.accessibility.keyboard_navigation.series_navigation import SeriesNavigation


class KeyboardNavigation(HighchartsMeta):
    """Options for Keyboard Navigation."""

    def __init__(self, **kwargs):
        self._enabled = None
        self._focus_border = None
        self._order = None
        self._series_navigation = None
        self._wrap_around = None

        self.enabled = kwargs.get('enabled', None)
        self.focus_border = kwargs.get('focus_border', None)
        self.order = kwargs.get('order', None)
        self.series_navigation = kwargs.get('series_navigation', None)
        self.wrap_around = kwargs.get('wrap_around', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'accessibility.keyboardNavigation'

    @property
    def enabled(self) -> Optional[bool]:
        """Enable keyboard navigation for the chart. Defaults to ``True``.

        :returns: Flag indicating whether keyboard navigation is enabled for the chart.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def focus_border(self) -> Optional[FocusBorder]:
        """Options for the focus border drawn around elements while navigating through
        them.

        :returns: Configuration of the focus border to be drawn around elements while
          navigating through them using the keyboard.
        :rtype: :class:`FocusBorder` or :obj:`None <python:None>`
        """
        return self._focus_border

    @focus_border.setter
    @class_sensitive(FocusBorder)
    def focus_border(self, value):
        self._focus_border = value

    @property
    def order(self) -> Optional[List[str]]:
        """Order of tab navigation in the chart.

        Determines which elements are tabbed to first. Available elements are:
          * ``'series'``
          * ``'zoom'``
          * ``'rangeSelector'``
          * ``'chartMenu'``
          * ``'legend'``
          * ``'container'``.

        In addition, any custom components can be added here. Adding container first in
        order will make the keyboard focus stop on the chart container first, requiring
        the user to tab again to enter the chart.

        :returns: Order of components to follow when tab navigating the chart.
        :rtype: :class:`list <python:list>` of :class:`str <python:str>` or
          :obj:`None <python:None>`
        """
        return self._order

    @order.setter
    def order(self, value):
        value = validators.iterable(value, allow_empty = True)
        if not value:
            value = None
        else:
            value = [validators.string(x, allow_empty = False) for x in value]

        self._order = value

    @property
    def series_navigation(self) -> Optional[SeriesNavigation]:
        """Options for the keyboard navigation of data points and series.

        :returns: Configuration for series navigation.
        :rtype: :class:`SeriesNavigation` or :obj:`None <python:None>`
        """
        return self._series_navigation

    @series_navigation.setter
    @class_sensitive(SeriesNavigation)
    def series_navigation(self, value):
        self._series_navigation = value

    @property
    def wrap_around(self) -> Optional[bool]:
        """If ``True``, indicates that keyboard navigation should wrap around when
        reaching the end of arrow-key navigation for an element in the chart.

        Defaults to ``True``.

        :returns: Flag indicating whether navigation should wrap around when reaching the
          end of arrow-key navigation for an element in the chart.
        :rtype: :class:`bool <python:bool>`:obj:`None <python:None>`
        """
        return self._wrap_around

    @wrap_around.setter
    def wrap_around(self, value):
        if value is None:
            self._wrap_around = None
        else:
            self._wrap_around = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        as_dict = validators.dict(as_dict, allow_empty = True) or {}
        kwargs = {
            'enabled': as_dict.get('enabled', None),
            'focus_border': as_dict.get('focusBorder', None),
            'order': as_dict.get('order', None),
            'series_navigation': as_dict.get('seriesNavigation', None),
            'wrap_around': as_dict.get('wrapAround', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'enabled': self.enabled,
            'focusBorder': self.focus_border,
            'order': self.order,
            'seriesNavigation': self.series_navigation,
            'wrapAround': self.wrap_around
        }
        return untrimmed


__all__ = [
    'KeyboardNavigation',
    'FocusBorder',
    'FocusBorderStyle',
    'SeriesNavigation'
]



from typing import Optional

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class AnnotationEvent(HighchartsMeta):
    """JavaScript callback functions that fire in response to annotation-related
    events."""

    def __init__(self, **kwargs):
        self._add = None
        self._after_update = None
        self._click = None
        self._remove = None

        self.add = kwargs.get('add', None)
        self.after_update = kwargs.get('after_update', None)
        self.click = kwargs.get('click', None)
        self.remove = kwargs.get('remove', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations.events'

    @property
    def add(self) -> Optional[CallbackFunction]:
        """JavaScript callback function called when an annotation is added to a chart.

        :returns: JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._add

    @add.setter
    @class_sensitive(CallbackFunction)
    def add(self, value):
        self._add = value

    @property
    def after_update(self) -> Optional[CallbackFunction]:
        """JavaScript callback function called when an annotation is updated (e.g. drag
        and dropped or resized by control points).

        :returns: JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._after_update

    @after_update.setter
    @class_sensitive(CallbackFunction)
    def after_update(self, value):
        self._after_update = value

    @property
    def click(self) -> Optional[CallbackFunction]:
        """JavaScript callback function called when an annotation is clicked.

        :returns: JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._click

    @click.setter
    @class_sensitive(CallbackFunction)
    def click(self, value):
        self._click = value

    @property
    def remove(self) -> Optional[CallbackFunction]:
        """JavaScript callback function called when an annotation is removed from the
        chart.

        :returns: JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._remove

    @remove.setter
    @class_sensitive(CallbackFunction)
    def remove(self, value):
        self._remove = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'add': as_dict.get('add', None),
            'after_update': as_dict.get('afterUpdate', None),
            'click': as_dict.get('click', None),
            'remove': as_dict.get('remove', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'add': self.add,
            'afterUpdate': self.after_update,
            'click': self.click,
            'remove': self.remove
        }

        return untrimmed



from typing import Optional, List

from highcharts_core.options.series.base import SeriesBase
from highcharts_core.options.series.data.cartesian import CartesianData
from highcharts_core.options.plot_options.spline import SplineOptions
from highcharts_core.utility_functions import mro__to_untrimmed_dict


class SplineSeries(SeriesBase, SplineOptions):
    """Options to configure a Spline series.

    A spline series is a special type of line series, where the segments between the
    data points are smoothed.

    .. figure:: ../../../_static/spline-example.png
      :alt: Spline Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def data(self) -> Optional[List[CartesianData]]:
        """Collection of data that represents the series. Defaults to
        :obj:`None <python:None>`.

        While the series type returns a collection of :class:`CartesianData` instances,
        it accepts as input three different types of data:

        .. tabs::

          .. tab:: 1D Collection

            .. code-block::

              series = SplineSeries()
              series.data = [0, 5, 3, 5]

            A one-dimensional collection of numerical values. Each member of the
            collection will be interpreted as a ``y``-value, with its corresponding ``x``
            value automatically determined.

            If :meth:`SplineSeries.point_start` is :obj:`None <python:None>`, ``x``
            values will begin at ``0``. Otherwise, they will start at ``point_start``.

            If :meth:`SplineSeries.point_interval` is :obj:`None <python:None>`, ``x``
            values will be incremented by ``1``. Otherwise, they will be incremented
            by the value of ``point_interval``.

          .. tab:: 2D Collection

            .. code-block::

              series = SplineSeries()
              # Category X-axis
              series.data = [
                  ['Category A', 0],
                  ['Category B', 5],
                  ['Category C', 3],
                  ['Category D', 5]
              ]

              # Numerical X-axis
              series.data = [
                  [9, 0],
                  [1, 5],
                  [2, 3],
                  [7, 5]
              ]

            A two-dimensional collection of values. Each member of the collection will be
            interpreted as an ``x`` and ``y`` pair. The ``x`` value can be a
            :class:`str <python:str>`, :class:`date <python:datetime.date>`,
            :class:`datetime <python:datetime.datetime>`, or numeric value.

            .. note::

              If the ``x`` value is a :class:`str <python:str>`, it will be interpreted
              as the name of the data point.

          .. tab:: Object Collection

            A one-dimensional collection of :class:`CartesianData` objects.

        :rtype: :class:`list <python:list>` of :class:`CartesianData` or
          :obj:`None <python:None>`
        """
        return self._data

    @data.setter
    def data(self, value):
        if not value:
            self._data = None
        else:
            self._data = CartesianData.from_array(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'data': as_dict.get('data', None),
            'id': as_dict.get('id', None),
            'index': as_dict.get('index', None),
            'legend_index': as_dict.get('legendIndex', None),
            'name': as_dict.get('name', None),
            'stack': as_dict.get('stack', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),
            'z_index': as_dict.get('zIndex', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = mro__to_untrimmed_dict(self, in_cls = in_cls)

        return untrimmed
